#ifndef DB_DRIVER_H
#define DB_DRIVER_H

#include <mysql.h>
#include <memory>
#include <string>

#include "user_workspace.h"

#define DB_CONF_BUF_SIZE 32

/**
 * @brief Container for configuration neccessary to open database connnection 
 * 
 */
typedef struct  {
    std::string     host;
    std::string     user;
    std::string     passwd;
    std::string     db;
    unsigned int    port;
    std::string     socket;
    unsigned long   clientflags;
} db_driver_conf;

/**
 * @brief Struct used to contain username, password and user status until credentials are verified
 * 
 */
typedef struct {
    char uname[64];
    // uint8_t passwd[64];
    char passwd[64];
    uint8_t status;
} user_cred; //user_credentials - container used to verify user credentials before retrieving all data

/**
 * @brief Handles database querries.
 * 
 */
class db_driver {
    public:
        /**
         * @brief Construct a new db driver object from configuration
         * 
         * @param conf
         */
        db_driver(db_driver_conf *conf);
        /**
         * @brief Construct a new db driver object through value passing. This is inteded for testing and 
         * debugging purposes
         */
        db_driver(const char* host = "localhost", const char* user = "pi", const char* passwd = "raspberry",
                  const char* db = "control_system", unsigned int port = 3306, const char* socket = NULL,
                  unsigned long cflags = 0);

        /** Checks if database is open */
        bool is_open();
        /**
         * @brief Selects username, password and status into @creds filtered by @uname
         * 
         * @return int 0 on success
         */
        int query_user_credentials(user_cred* creds, const char* uname);
        /**
         * @brief Selects all user data into user container @usr. Data is selected by either username
         * or rfid_serial. If both are provided rfid takes precedence
         * 
         * @return int 0 on success
         */
        int query_user_data(user_cont* usr, const char* uname = nullptr,
                            uint8_t* rfid_serial = nullptr, uint8_t rfid_ser_len = 0);
        /** Selects username into @uname for user with id = @id */
        int query_username(std::string& uname, unsigned long id);
        /** Selects basic information about measurements, that is initialized into @usr cont variable for headers.
         * These headers are used to list users measurements
         */
        int query_measurement_headers(user_cont* usr);
        /* Selects all measurement data into @m by @id */
        int query_measurement(measurement* m, unsigned long id);
        /** queries continuous measurements (they have same measurement number but different measurement_part)
         *  every user's measurement number should be unique. Don't select all the info that query_measurement does. */
        int query_continuous_measurement(std::vector<measurement> &ms, unsigned long m_number,
            unsigned long measuree_id);
        /** Queries the database to see if measurement wiht number @m_number of user with id @masuree_id. 
         * Returns 1 if the measuring is continous, 9 if its a single measuring, other values otherwise
         */
        int is_measurement_continuous(unsigned long m_number, unsigned long measuree_id);
        /** Queries the database to see if @uname is available for new user registration. 
         * Returns 0 if name is available */
        int is_username_available(const char* uname);
        /** Queries the database to see if rfid @serial (length @ser_len) is available for new user registration. 
        * Returns 0 if name is available */
        int is_rfid_serial_available(uint8_t *serial, uint8_t ser_len);
        /** Increments user measrue count after a measurement is successfully performed for user with @id */
        int increment_user_measurement_count(unsigned long id);
        /** Checks if user with @uname is a pacient, so he can be logged as subuser. Returns 0 for true. */
        int can_login_subuser(const char* uname);
        /** Updates users @user_id rfid card with @serial (len @ser_len). However this is currently not used.
         * Returns 0 on success.
        */
        int update_user_rfid(unsigned long user_id, uint8_t* serial, uint8_t ser_len);
        /** Inserts new user into the database with data read from @new_user and @password (password isn't stored in 
         * user container) Returns 0 on success.*/
        int insert_user(user_cont *new_usr, const char* password);
        /** Inserts measurement with data read from @m. Returns 0 on success. */
        int insert_measurement(measurement *m);

        /** Test insert with binary data */
        void test_insert_with_binary();
        /** Test select with binary data */
        void test_select_with_binary();

    private:
        /** flag that is set to true if connection is successfully made */
        bool db_is_open = true;
        /** Established connection */
        std::unique_ptr<MYSQL> conn;

        int test(); // this shouldn't be used anywhere, non prepared satement test
        /** When a query failes, this function logs out error and closes statement @stmt */
        void handle_stmt_error(MYSQL_STMT* stmt);
        /** Converts MYSQL_TIME structre @src to C++ tm structure @dest */
        void mysql_time_to_tm(MYSQL_TIME* src, std::tm *dest);
        // int prprd_stmt_executor(const char *sql, const char* fnc_name, MYSQL_BIND *par, int par_cnt,
        //      MYSQL_BIND *res, int res_cnt);
};

#endif