#ifndef USER_WORKSPACE_H
#define USER_WORKSPACE_H

#include <memory>
#include <string>
#include <ctime>
#include <vector>
#include "rfid_reader.h"
#include "hx711/common.h"

/**
 * @brief Serves as container for basic measurement information, that is dispayed to GUI.
 * 
 */
class measurement_header {
    public:
        unsigned long id;
        unsigned long measurement_number;
        unsigned long measuree_id;
        std::tm measuring_start;

        const char* get_header_label();
    
    private:
        std::string label;
};

class user_workspace; // defined later, due to changes its required in measurement

/**
 * @brief Container for measurement data and (convenience) variables, that are used to show measurement in GUI
 * 
 */
class measurement {
    public:
        // this follows db structure
        unsigned long id;
        // uint8_t status; // isn't selected from db and isn't used anywhere anyway
        unsigned long measurement_number;
        unsigned long measurement_part = 1;
        unsigned long measuree_id;
        unsigned long measurer_id;
        unsigned long measuring_length; // in microsec (us)
        std::tm measuring_start;
        std::tm measuring_end;
        std::string description;
        // std::string unit; // not required, raw numbers don't need to know unit to convert
        uint8_t value_length;
        unsigned long value_count;
        std::vector<double> values; // in grams, only available for single measurements
        // double raw_avg;  // redundant - same as all_val_avg
        double x_val_med; // in grams
        double x_val_avg; // in grams
        double all_val_med; // in grams
        double all_val_avg; // in grams
        std::tm created_at;

        // continuous
        bool continuous = false;
        int cont_measurements; // number of measurements (parts) taken during continous measurement


        // convinience vars - these are used in GUI
        HX711::Mass::Unit current_unit = HX711::Mass::Unit::G;
        std::string x_val_med_str;
        std::string x_val_avg_str;
        std::string all_val_med_str;
        std::string all_val_avg_str;
        std::string measuree_uname;
        std::string measurer_uname;
        std::string length_str;
        std::string start_str;
        std::string end_str;


        /** Simple measuring is a measuring with values only. This is used for the values to be observed during measuring */
        static void init_simple_measuring(measurement *m, std::vector<double> &values);
        /** Properly inits measuring values, that are inserted into database */
        static void init_measuring(measurement *m, user_workspace *uspace, std::vector<double> &values,
            std::tm *start_tm, std::tm *end_tm, unsigned long m_length);
        /** Calculates Avarage and Median of measuring */
        static void calc_measuring_results(measurement *m);
        /** Calculates Avarages of Avarages and Medians of several measurings, that are then put into @picked as a single measuring */
        static void prepare_continuous_to_picked(std::vector<measurement> &ms, measurement* picked);
        
        /**
         * @brief Calculates the median of medians and avarages of avarages of every X values 
         *  (X is a defined constant in .cpp). Is maybe avarage of medians preferred (can be changed)?
         * 
         * @return std::pair<double, double>  - first is avarage of medians, second is avarage of avarages
         */
        static std::pair<double, double> calc_every_x_values_stats(measurement *m);
        
        /**
         * @brief Calculates the median and avarage of all values
         * 
         * @return std::pair<double, double>  first is median, second is avarage
         */
        static std::pair<double, double> calc_all_values_stats(measurement *m);

        /** Prints measuring on debug level */
        void log_measurement_to_debug();
        /** Intializes numbers to string, that can be displayed in GUI */
        void init_value_strings();
        /** Inits strings for start and end date/times */
        void init_start_end_strings();
        /** Calls @init_value_strings and @init_start_end_strings and inits usernames with parameters @measurer_uname and @measuree_uname */
        void init_convinience_vars(const char* measurer_uname, const char* measuree_uname);
        /** Compares @this to @m and prints out non-matching values */
        int compare(measurement* m);
};

/**
 * @brief Class used as a container for user data, that are selected from database. (and also for insert,
 * but not everything is used in that case)
 * 
 */
class user_cont {
    public: 
        unsigned long id;
        unsigned long measure_count;
        std::string rfid_serial;
        uint8_t rfid_serial_bin[10];
        uint8_t rfid_ser_len;
        std::string username;
        std::string name;
        std::string lastname;
        std::string description;
        uint8_t role;
        std::tm reg_date;
        std::tm last_login_date;
        std::tm created_at_date;
        std::tm date_of_birth;

        /** Vector of headers, that are displayed on screen 3 and user can select them */
        std::vector<std::unique_ptr<measurement_header>> measur_headers;

        /**
         * @brief Construct a new user cont object through value parameters, but this isn't used. Originally inteded 
         * for testing purposes
         */
        user_cont(unsigned long, uint8_t*, uint8_t, std::string, std::string, std::string,
                  std::string, uint8_t, std::tm, std::tm, std::tm, std::tm);
        /** Prints user on debug level */
        void log_user_to_debug();
}; // user_container, not using namespace hence user_cont as name instead of just "user"

/**
 * @brief Container that servers for storing users data and manipulating them.
 * 
 */
class user_workspace {
    public:
        // user loaded by user_logged (emp. can login pac. under them to measure/ view result)
        std::unique_ptr<user_cont> subuser;
        std::unique_ptr<measurement> measured;
        std::unique_ptr<measurement> picked;

        std::vector<measurement> continuous_m;
        std::unique_ptr<measurement> continuous_front;

        /**
         * @brief Construct a new user workspace object. This can only be done with a user that is logged in.
         * 
         * @param user 
         */
        user_workspace(user_cont* user);

        /** Inits measuring into @measured attribute */
        void init_measured(std::vector<double> values, std::tm *start_tm, std::tm *end_tm, unsigned long m_length);
        /** Returns @measured */
        measurement *get_last_hx_measuring();
        /** Returns @picked (last selected measuring on screen 3) */
        measurement *get_picked_hx_measuring();
        /** Compere RFID serial of @user to @tag serial */
        int compare_rfid_serials(rfid_reader::rfid_tag *tag);

        bool has_user(); /** Has user getter */
        bool has_subuser(); /** Has subuser getter */
        std::pair<uint8_t*, uint8_t> get_rfid_serial(); /** Gets RFID serial (with length) */
        unsigned long get_user_id(); /** user id getter */
        user_cont* get_user(); /** Returns user container */
        std::string& get_username(); /** get users's username */
        std::string& get_name(); /** get user's name */
        std::string& get_lastname(); /** get user's last name */
        std::string& get_description(); /** get user's description */
        uint8_t get_role(); /** get user's role */
        const char* get_role_string(); /** get user's role as string to show in GUI */
        /** get reference to measurement headers */
        std::vector<std::unique_ptr<measurement_header>>& get_user_measur_headers();

        user_cont* get_measured_user(); /** returns either subuser or user */
        unsigned long get_measured_id(); /** returns either subuser id or user id */
        std::string& get_measured_username(); /** returns either subuser username or user username */

    private:
        std::unique_ptr<user_cont> user;
};


#endif