#include "screen_definitions.h"

#include <spdlog/spdlog.h>

#include "gui_button.h"
#include "gui_combobox.h"
#include "gui_direct.h"
#include "gui_input.h"
#include "gui_label.h"
#include "gui_observer.h"
#include "gui_selectable.h"
#include "localisation.h"
#include "app_workspace.h"
#include "config_loader.h"

// limit definitions moved to app_workspace.h so they can be used here and in app_workspace

#define EMPTY_LINE elems.push_back(std::unique_ptr<gui_element>(new gui_direct(empty_line)));
#define SAME_LINE  elems.push_back(std::unique_ptr<gui_element>(new gui_direct(same_line)));
#define INDENT     elems.push_back(std::unique_ptr<gui_element>(new gui_direct(indent)));
#define UNINDENT   elems.push_back(std::unique_ptr<gui_element>(new gui_direct(unindent)));
#define INSERT_ELEMENT(ELEMENT) elems.push_back(std::unique_ptr<gui_element>(ELEMENT));

/*********************************************************************************************************************/
/*       Local function definitions (used in screens)                                                                */
/*********************************************************************************************************************/

gui_element* last_element(std::vector<std::unique_ptr<gui_element>> &elems);
void empty_line();
void same_line();
// void begin_child_window(); /** Creates a "subwindow" in current window. Used for scrollable region */
// void end_child_window();
void begin_scrollabe_region(); /** Creates region without nav */
void end_scrollabe_region();

static float custom_indent = 20.0f; // for customizing identation if needed
/** Indents all elements after this by custom_indent value. To cancel indentation "unindent" must be called
 *  For simpler use call macros: INDENT/ UNINDENT
 */
void indent();
void unindent();

// screen_1
void login_action();
void login_subuser();
void logout_action();
void logout_subuser();

// screen_2
void start_hx_measuring();
void start_hx_continuous();

// screen_3
void open_measuring_detail();
void apply_meas_list_filter();

// screen_4
void unit_select_action();

// screen_5
void move_to_hx_calibration();
void move_to_user_addition();
void move_to_hx_test_samp();
void move_to_hx_test_time();
void move_to_kb_delay_test();

// screen_6
void start_tare();
void start_calibration();
void apply_calibration();
void register_user();

/** debug **/

// leftovers - remove ?
void cb();
void draw_rect();
void render_line();
void render_vert();

// actual debug
void empty_cb();
void callback_debug();
void int_test_label();
void float_test_label();
void text_test_label();
void text_wrap_debug();

// db test calls
void invoke_sql_debug();
void test_user_login_sql();
void test_user_creds_sql();
void test_user_meas_header_sql();
void test_user_inc_measur_cnt();
void test_user_is_uname_avail();
void test_user_is_rfid_avail();
void test_user_insert();
void test_measurement_insert();
void test_measurement_query();
void update_rfids_debug();

void hx_samples_debug();
void hx_timeout_debug();

void test_errscr_1();
void test_errscr_2();


/*********************************************************************************************************************/
/*       Implementation of header function definitions (screen definitions)                                          */
/*                                                                                                                   */
/*       Used GUI library uses "vertical" layout by default but offers options to specify position and width         */
/*       most widgets. Height is usually determined automatically from font size and settable vertical spacing.      */
/*       Through adding gui_elements to elems (vector of screen elements) a screen is defined with simplified API    */
/*                                                                                                                   */
/*       The way the library works is that "attributes" (like position/ width/ spacing...) are specified             */
/*       for the following elements. For example programmer using this API doesn't have to use constructor           */
/*       with position, but can specify position and other attributes in a function and call it through gui_direct   */
/*       and then add element like gui_label which will have attributes from previous gui_direct                     */
/*********************************************************************************************************************/

gui_label *err_title = new gui_label("Changable error title");
gui_label *err_desc = new gui_label("Changable error description");

void define_screen_0(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_0_ERR_TITLE_LABEL")))
    INDENT
    INSERT_ELEMENT(err_title)
    UNINDENT
    EMPTY_LINE
    INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_0_ERR_DESC_LABEL")))
    INDENT
    INSERT_ELEMENT(err_desc)
    UNINDENT
}

void define_screen_1(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    // if (usr_wrk != nullptr) { //get_userspace() checks if user exists = is logged in
    if (app_workspace::get_instance()->has_user()) {
        user_workspace* usr_wrk = app_workspace::get_instance()->get_userspace();

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_1_LOGGED_USERNAME")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(usr_wrk->get_username().c_str()))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_1_LOGGED_ROLE")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(usr_wrk->get_role_string()))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_1_LOGOUT_BTN"), logout_action, -1, -1, 460, 40))

        if (usr_wrk->get_role() >= 1) { // emp/ admin
            if (usr_wrk->subuser.get() != nullptr) {
                EMPTY_LINE
                EMPTY_LINE
                INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_1_LOGGED_SUBUSERNAME")))
                SAME_LINE
                INSERT_ELEMENT(new gui_label(usr_wrk->subuser->username.c_str()))
                INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_1_LOGOUT_SUBUSER"),
                    logout_subuser, -1, -1, 460, 40))

            } else {
                EMPTY_LINE
                EMPTY_LINE
                INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_1_LOGIN_SUBUSER")))
                INSERT_ELEMENT(new gui_input(IN_TEXT, (void*) &app_workspace::get_instance()->subusr_name,
                            sizeof(char) * DEF_BUFF_SIZE, -1 ,-1 ,-1, get_localized_text("SCREEN_1_USERNAME")))
                INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_1_LOGIN_SUBUSER_BTN"), 
                            login_subuser, -1, -1, 460, 40)) // TDOO change action to login_subuser
            }
        }
    } else {
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_1_LOGIN_W_RFID")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_OR")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_input(IN_TEXT, (void*) &app_workspace::get_instance()->usr_name,
                        sizeof(char) * DEF_BUFF_SIZE, -1 ,-1 ,-1, get_localized_text("SCREEN_1_USERNAME")))
        INSERT_ELEMENT(new gui_input(IN_PASSWORD, (void*) &app_workspace::get_instance()->usr_passwd,
                        sizeof(char) * DEF_BUFF_SIZE_SMALL, -1, -1, -1, get_localized_text("SCREEN_1_PASSWORD")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_1_LOGIN_BTN"), login_action, -1, -1, 460, 40))
    }
}

void define_screen_2(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();
    app_workspace *app_wrk = app_workspace::get_instance().get();
    user_workspace *usr_wrk = app_wrk->get_userspace();

    if (!app_wrk->has_user()) {
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_SCREEN_REQUIRE_USER")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_FUNCTION_DESC")))
        return;
    }

    if (app_wrk->hx_measuring) {
        if (app_wrk->hx_not_finished_yet) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_WAITING_FOR_FINISH")))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING_CANCEL_ENTER")))
            return;
        }

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING")))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING_CANCEL_ENTER")))
        EMPTY_LINE

        measurement *m = app_wrk->observed_measurement.get();
        if (m != nullptr) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_MEDIAN"), -1, -1, 250))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->x_val_med_str.c_str()))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_AVARAGE"), -1, -1, 250))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->x_val_avg_str.c_str()))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEDIAN")))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->all_val_med_str.c_str()))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_AVARAGE")))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->all_val_avg_str.c_str()))
        }

        // if (app_workspace::get_instance()->hx_continuous) {
        //     INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING")))
        //     INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING_CANCEL_ENTER")))
        // } else {
        //     INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING")))
        //     INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING_WAIT_PLEASE")))
        // }
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_CHANGED_MEASURED_USER_TIP"), -1, -1, -1,
                                     app_workspace_ns::font_size::SMALL_FONT))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURED_USER")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(usr_wrk != nullptr ? 
                                     usr_wrk->get_measured_username().c_str() :
                                     get_localized_text("GT_USER_NO_LOGIN")))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEASURING_OPTIONS")))

        INDENT

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_CB1_LABEL"), -1, -1, 160))
        SAME_LINE
        INSERT_ELEMENT(new gui_combobox(app_wrk->hx_measure_opts, S2_MEASURE_OTPS_CNT,
                                        &(app_wrk->selected_hx_measure_opt), nullptr, 160, -1, 200))
        last_element(elems)->set_refresh_screen(&(app_wrk->refresh_current_screen));

        if (app_wrk->selected_hx_measure_opt == 0) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_SAMPLES_IN_LIMITS"), -1, -1, -1,
                                         app_workspace_ns::font_size::SMALL_FONT))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_CB1_SAMPLES"), -1, -1, 160))
            SAME_LINE
            INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->samples_in),
                                         0, 160, -1, 200))
            ((gui_input*) last_element(elems))->set_min_max(S2_MIN_SAMPLES_IN, S2_MAX_SAMPLES_IN); // WARNING! - if used after wrong element, application crashes on startup
        } else if (app_wrk->selected_hx_measure_opt == 1) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_TIMEOUT_IN_LIMITS"), -1, -1, -1,
                                         app_workspace_ns::font_size::SMALL_FONT))
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_CB1_TIMEOUT"), -1, -1, 160))
            SAME_LINE
            INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->timeout_in),
                                         0, 160, -1, 200))
            ((gui_input*) last_element(elems))->set_min_max(S2_MIN_TIMEOUT_IN, S2_MAX_TIMEOUT_IN); // WARNING! - if used after wrong element, application crashes on startup
            SAME_LINE
            // INSERT_ELEMENT(new gui_label(get_localized_text("GT_SEC_UNIT")))
            INSERT_ELEMENT(new gui_label(get_localized_text("GT_SEC_UNIT_BRTS")))
        }

        UNINDENT

        EMPTY_LINE
        INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_2_START_MEASURING"), 
                                      start_hx_measuring, -1, -1, 460, 40))
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_OR")))
        INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_2_START_CONTINOUS"), 
                                      start_hx_continuous, -1, -1, 460, 40))                                      

        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_LAST_MEAS_RES")))
        if ((app_wrk->has_user() && app_wrk->userspace->get_last_hx_measuring() != nullptr &&
            app_wrk->userspace->get_last_hx_measuring()->values.size() > 0))
        {
            measurement *m = app_wrk->userspace->get_last_hx_measuring();
            // m->init_value_strings(); // this should be init after measuring

            // allow unit selection on this screen? or statically one unit and unit selection on measuring detail
            // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_CB2_LABEL"), -1, -1, 160))
            // SAME_LINE
            // INSERT_ELEMENT(new gui_combobox(result_unit_opts, 2, &(app_wrk->selected_hx_result_unit),
                                            // nullptr, 160, -1, 200))
            // last_element(elems)->set_refresh_screen(&(app_wrk->refresh_current_screen));

            INDENT

            // raw_avg is actually the same as all_val_avg
            // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_RAW_AVG")))
            // SAME_LINE
            // INSERT_ELEMENT(new gui_label("result"))
            // INSERT_ELEMENT(new gui_observer(OBS_FLOAT, &(m->raw_avg))); //gui_observer can serve as label for numerical values
            // SAME_LINE

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_MEDIAN"), -1, -1, 250))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->x_val_med_str.c_str()))

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_AVARAGE"), -1, -1, 250))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->x_val_avg_str.c_str()))

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEDIAN")))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->all_val_med_str.c_str()))

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_AVARAGE")))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(m->all_val_avg_str.c_str()))

            UNINDENT

        } else {
            INDENT
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_NO_LAST_RESULT")))
            UNINDENT
        }
    }
}

void define_screen_3(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    app_workspace *app_wrk = app_workspace::get_instance().get();

    if (app_workspace::get_instance()->has_user()) {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_3_USER_MEASURINGS")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(app_workspace::get_instance()->userspace->get_measured_username().c_str()))

        std::vector<std::unique_ptr<measurement_header>> &headers = 
            app_workspace::get_instance()->userspace->get_user_measur_headers();
        
        if (headers.size() <= 0) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_3_MEASUREMENTS_EMPTY")))
        } else {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_3_FILTER_LABEL")))

            INSERT_ELEMENT(new gui_combobox(app_wrk->s3_filter_opt_labels, S3_FILTER_OPT_CNT,
                                        &(app_wrk->s3_selected_filter), nullptr, -1, -1, 250))
            ((gui_combobox*) last_element(elems))->set_value_change_action(apply_meas_list_filter); // WARNING! - if used after wrong element, application crashes (due to conversion)
            SAME_LINE
            INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->s3_filter_number), 0))
            // set min filter value to 0, because measurement_number can only be unsigned
            ((gui_input*) last_element(elems))->set_min_max(S3_MIN_FILTER_NUMBER); // WARNING! - if used after wrong element, application crashes on startup
            ((gui_input*) last_element(elems))->set_value_change_action(apply_meas_list_filter); // WARNING! - if used after wrong element, application crashes (due to conversion)

            INSERT_ELEMENT(new gui_direct(begin_scrollabe_region))

            if (app_wrk->s3_selected_filter == 0) { // Larger than
                for (size_t i = 0; i < headers.size(); i++) {
                    measurement_header *h =  headers.at(i).get();

                    if (h->measurement_number > app_wrk->s3_filter_number) {
                        INSERT_ELEMENT(new gui_selectable(h->get_header_label(), i, 
                            &(app_workspace::get_instance()->selected_measuring), open_measuring_detail));
                    }
                }
            } else if (app_wrk->s3_selected_filter == 1) { // smaller than
                for (size_t i = 0; i < headers.size(); i++) {
                    measurement_header *h =  headers.at(i).get();

                    if (h->measurement_number < app_wrk->s3_filter_number) {
                        INSERT_ELEMENT(new gui_selectable(h->get_header_label(), i, 
                            &(app_workspace::get_instance()->selected_measuring), open_measuring_detail));
                    }
                }
            } else {
                for (size_t i = 0; i < headers.size(); i++) {
                    measurement_header *h =  headers.at(i).get();
                    INSERT_ELEMENT(new gui_selectable(h->get_header_label(), i, 
                        &(app_workspace::get_instance()->selected_measuring), open_measuring_detail));
                }
            }

            INSERT_ELEMENT(new gui_direct(end_scrollabe_region))
        }
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_SCREEN_REQUIRE_USER")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_3_FUNCTION_DESC")))
    }
}

void define_screen_4(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    app_workspace* app_wrk = app_workspace::get_instance().get();
    measurement* m = nullptr;

    if (app_wrk->has_user()) {
        m = app_wrk->userspace->picked.get();
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_SCREEN_REQUIRE_USER")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_FUNCTION_DESCRIPTION")))
        return;
    }

    INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_TITLE")))

    if (m == nullptr || m->id <= 0) {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_NO_MEASUREMENT")))
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_UNITS_LAB"), -1, -1, 180))
        SAME_LINE
        INSERT_ELEMENT(new gui_combobox(app_wrk->unit_sel_labels, UNIT_SEL_CNT,
                                        &(app_wrk->selected_unit_option), nullptr, 160, -1, 200))
        ((gui_combobox*) last_element(elems))->set_value_change_action(unit_select_action); // WARNING! - if used after wrong element, application crashes (due to conversion)

        // using gui_observer for numerical values, because label would require some buffer/ string

        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASUR_NUM_LAB"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASUR_NUM_LAB")))
        SAME_LINE
        INSERT_ELEMENT(new gui_observer(OBS_LONG, &(m->measurement_number)))

        if (m->continuous) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASUR_NUM_PARTS")))
            SAME_LINE
            INSERT_ELEMENT(new gui_observer(OBS_INT, &(m->cont_measurements)))
        }

        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURED_LAB"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURED_LAB")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->measuree_uname.size() <= 0 ?
                        get_localized_text("GT_EMPTY_NULL_STRING") : m->measuree_uname.c_str()))

        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURED_BY_LAB"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURED_BY_LAB")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->measurer_uname.size() <= 0 ?
                        get_localized_text("GT_EMPTY_NULL_STRING") : m->measurer_uname.c_str()))


        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURING_SECTION")))
        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURING_LEN_LAB"), -1, -1, 250))
        INDENT
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_M_LENGTH"), -1, -1, 150))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->length_str.c_str()))

        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURING_START_LAB"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_M_START"), -1, -1, 150))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->start_str.c_str()))

        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_MEASURING_END_LAB"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_M_END"), -1, -1, 150))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->end_str.c_str()))
        UNINDENT

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_EVERY_X_VAL_SECTION")))
        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_MEDIAN"), -1, -1, 250))
        INDENT
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEDIAN"), -1, -1, 150))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->x_val_med_str.c_str()))
        
        // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_X_VAL_AVARAGE"), -1, -1, 250))
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_AVARAGE"), -1, -1, 150))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->x_val_avg_str.c_str()))
        UNINDENT
        

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_ALL_VAL_SECTION")))
        INDENT
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_MEDIAN")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->all_val_med_str.c_str()))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_AVARAGE")))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(m->all_val_avg_str.c_str()))
        UNINDENT
    }
}

void define_screen_5(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    app_workspace* app_wrk = app_workspace::get_instance().get();

    if (app_wrk->s5_success_indicator >= 0) {
        if (app_wrk->s5_success_indicator == S5_CALIBRATION_APPLIED) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_CALIBRATION_APPLIED")))
        } else if (app_wrk->s5_success_indicator == S5_USER_CREATED) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_USER_CREATED")))
        } else if (app_wrk->s5_success_indicator == S5_FINISHED_KB_DELAYS) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_FINISHED_TESTING_KB")))
        }
    }

    if (app_wrk->has_user()) {
        if (app_wrk->userspace->get_user()->role <= 1) {
            INSERT_ELEMENT(new gui_label(get_localized_text("GT_INSUFFICIENT_PERMISSION")))
            EMPTY_LINE
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_ADMIN_SCREEN_ONLY")))
            return;
        }
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_SCREEN_REQUIRE_USER")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_FUNCTION_DESCRIPTION")))
        return;
    }

    INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_5_CALIBRATE_HX"), 
                                  move_to_hx_calibration, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_5_ADD_USER"), 
                                  move_to_user_addition, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_5_TEST_HX_SAMPLES"), 
                                  move_to_hx_test_samp, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_5_TEST_HX_TIMEOUT"), 
                                  move_to_hx_test_time, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_5_TEST_KB_DELAYS"), 
                                  move_to_kb_delay_test, -1, -1, 460, 40))
}

void define_screen_6(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    app_workspace* app_wrk = app_workspace::get_instance().get();

    if (app_wrk->has_user()) {
        if (app_wrk->userspace->get_user()->role <= 1) {
            INSERT_ELEMENT(new gui_label(get_localized_text("GT_INSUFFICIENT_PERMISSION")))
            EMPTY_LINE
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_5_ADMIN_SCREEN_ONLY")))
            return;
        }
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_SCREEN_REQUIRE_USER")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_FUNCTION_DESCRIPTION")))
        return;
    }

    if (app_wrk->hx_measuring) {
        if (app_wrk->s6_tare_complete) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_CALIBRATING")))
        } else if (app_wrk->test_hx_testing) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_TESTING_IN_PROGRESS")))
        } else {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_TARING")))
        } 

        return;
    }

    if (app_wrk->s5_screen_6_indicator == S5_TO_S6_CALIBRATION) { // calibration
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_REF_UNIT_LAB"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->s6_ref_mass_in), 0, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_4_UNITS_LAB"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_combobox(app_wrk->s6_unit_sel_labels, S6_UNIT_SEL_CNT,
                                        &(app_wrk->s6_selected_unit_option), nullptr, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_SAMPLES_LAB"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->s6_samples_in), 0, 220, -1, 200))
        ((gui_input*) last_element(elems))->set_min_max(S6_MIN_SAMPLES_IN, S6_MAX_SAMPLES_IN); // WARNING! - if used after wrong element, application crashes on startup

        EMPTY_LINE

        if (app_wrk->s6_tare_complete) {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_CALIBRATION_REQ")))
            INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_6_START_CALIBRATION"), 
                                          start_calibration, -1, -1, 460, 40))
        } else {
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_TARE_REQ")))
            INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_6_START_TARE"), 
                                          start_tare, -1, -1, 460, 40))
        }

        if (app_wrk->s6_calibration_finished) {
            EMPTY_LINE
            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_REF_UNIT_LAB"), -1, -1, 240))
            SAME_LINE
            INSERT_ELEMENT(new gui_label(app_wrk->s6_ref_mass_str.c_str()))

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_HX_REF_UNIT_LAB"), -1, -1, 240))
            SAME_LINE
            INSERT_ELEMENT(new gui_observer(OBS_INT, &(app_wrk->s6_calib_ref_unit)))

            INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_HX_OFFSET_LAB"), -1, -1, 240))
            SAME_LINE
            INSERT_ELEMENT(new gui_observer(OBS_INT, &(app_wrk->s6_calib_offset)))

            EMPTY_LINE
            INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_6_APPLY_CAL"), 
                                          apply_calibration, -1, -1, 460, 40))

            // EMPTY_LINE
            // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_MAKE_PERMANENT_LAB")))
        }
    } else if (app_wrk->s5_screen_6_indicator == S5_TO_S6_USERADD) { // user addition
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_ROLE"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_combobox(app_wrk->s6_role_labels, S6_ROLE_CNT,
                                        &(app_wrk->s6_selected_role), nullptr, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_USERNAME"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_TEXT, &(app_wrk->s6_uname_in), DEF_BUFF_SIZE_SMALL, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_PASSWORD"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_PASSWORD_PLAIN, &(app_wrk->s6_passwd_in),
                                     DEF_BUFF_SIZE_SMALL, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_RFID_IND"), -1, -1, 280))
        SAME_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text(app_wrk->s6_rfid_indicator ?
                                                        get_localized_text("GT_YES") :
                                                        get_localized_text("GT_NO")), -1, -1, 220))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_DESCRIPTION"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_TEXT, &(app_wrk->s6_desc_in), DEF_BUFF_SIZE_EXTRA, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_NAME"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_TEXT, &(app_wrk->s6_name_in), DEF_BUFF_SIZE, 220, -1, 200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_LASTNAME"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_TEXT, &(app_wrk->s6_lastname_in), DEF_BUFF_SIZE_SMALL, 220, -1,200))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_DOB"), -1, -1, 300))

        INDENT

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_YEAR"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->s6_year_in), 0, 220, -1, 200))
        std::time_t ct_tmp = std::time(nullptr);
        std::tm *current_time = std::localtime(&ct_tmp);
        int year_max = current_time->tm_year + 1900; // tm_year is a year since 1900 (e.g: 122 = 2022)
        ((gui_input*) last_element(elems))->set_min_max(YEAR_MIN, year_max); // WARNING! - if used after wrong element, application crashes on startup


        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_MONTH"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_combobox(app_wrk->s6_month_labels, S6_MONTH_CNT,
                                        &(app_wrk->s6_selected_month), nullptr, 220, -1, 200))       

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_USR_DAY"), -1, -1, 220))
        SAME_LINE
        INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_wrk->s6_day_in), 0, 220, -1, 200))
        ((gui_input*) last_element(elems))->set_min_max(DAY_MIN, DAY_MAX); // WARNING! - if used after wrong element, application crashes on startup

        UNINDENT

        EMPTY_LINE
        INSERT_ELEMENT(new gui_button(get_localized_text("SCREEN_6_USR_ADD"), 
                                      register_user, -1, -1, 460, 40))
    } else if (app_wrk->s5_screen_6_indicator == S5_TO_S6_HX_TEST_SAMPLES) { // testing hx with sample count
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_RESULTS")))
        EMPTY_LINE
        
        for (int i = 0; i < (HX_TEST_CONST_IT_COUNT + HX_TEST_INC_IT_COUNT); i++) {
            INSERT_ELEMENT(new gui_label(app_wrk->test_hx_samples_times[i].second.c_str()))
        }
    } else if (app_wrk->s5_screen_6_indicator == S5_TO_S6_HX_TEST_TIMEOUT) { // testing hx with timeout
        INSERT_ELEMENT(new gui_label(get_localized_text("GT_RESULTS")))
        EMPTY_LINE
        
        for (int i = 0; i < (HX_TEST_CONST_IT_COUNT + HX_TEST_INC_IT_COUNT); i++) {
            INSERT_ELEMENT(new gui_label(app_wrk->test_hx_timeout_collected[i].second.c_str()))
        }
    } else if (app_wrk->s5_screen_6_indicator == S5_TO_S6_KB_DELAY_TEST) {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_KB_TESTING_TIP1")))
        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_KB_TESTING_TIP")))

        EMPTY_LINE
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_STOP_TESTING_HASH")))
        EMPTY_LINE

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_KB_DELAY_MS")))
        SAME_LINE
        INSERT_ELEMENT(new gui_observer(OBS_INT, &(app_wrk->kb_delay_ms)))

        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_KB_DELAY_US")))
        SAME_LINE
        INSERT_ELEMENT(new gui_observer(OBS_INT, &(app_wrk->kb_delay_us)))
    } else {
        INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_6_WRONG_INDICATOR")))
    }
}

///////////////////// debug screen definitons

void define_screen_1_debug(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    INSERT_ELEMENT(new gui_label("Inputs test", -1, -1, -1, app_workspace_ns::font_size::BIG_FONT))
    INSERT_ELEMENT(new gui_label("Int input"))
    // for non-text inputs "data_size" doesn't matter
    INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_workspace::get_instance()->int_test), 0))
    INSERT_ELEMENT(new gui_label("Float input"))
    // for non-text inputs "data_size" doesn't matter
    INSERT_ELEMENT(new gui_input(input_type::IN_FLOAT, &(app_workspace::get_instance()->float_test), 0))
    INSERT_ELEMENT(new gui_label("Text input"))
    INSERT_ELEMENT(new gui_input(input_type::IN_TEXT, &(app_workspace::get_instance()->text_test), DEF_BUFF_SIZE))
    INSERT_ELEMENT(new gui_label("Password input"))
    INSERT_ELEMENT(new gui_input(input_type::IN_PASSWORD, &(app_workspace::get_instance()->pass_test), DEF_BUFF_SIZE))
    EMPTY_LINE
    INSERT_ELEMENT(new gui_direct(int_test_label))
    INSERT_ELEMENT(new gui_direct(float_test_label))
    INSERT_ELEMENT(new gui_direct(text_test_label))

    INSERT_ELEMENT(new gui_label("Following line is defined with absolute positions and width (that's why the position"
            " is misaligned to other elements", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_label("Int input:", 20, 500, 200))
    INSERT_ELEMENT(new gui_input(input_type::IN_INT, &(app_workspace::get_instance()->int_test), 0, 200, 500, 100))
    INSERT_ELEMENT(new gui_label("This should be on next line", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    // newly added gui_observer -test
    INSERT_ELEMENT(new gui_label("Int observable: ", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    SAME_LINE
    INSERT_ELEMENT(new gui_observer(OBS_INT, &(app_workspace::get_instance()->int_test), -1, -1, -1,
                                    app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_label("Float observable: ", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    SAME_LINE
    INSERT_ELEMENT(new gui_observer(OBS_FLOAT, &(app_workspace::get_instance()->float_test), -1, -1, -1,
                                    app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_label("String observable: ", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    SAME_LINE
    INSERT_ELEMENT(new gui_observer(OBS_STRING, app_workspace::get_instance()->text_test, -1, -1, -1,
                                    app_workspace_ns::font_size::SMALL_FONT))

    /* // testing const char* as parameter instead char[]
    static const char* test = "Test string";
    elems.push_back(std::unique_ptr<gui_element>(
        new gui_observer(OBS_STRING, (void*) test)
    ));
    */
}


// variables used in debug callback in screen_2_debug
static bool cb_flag = false;
static int cb_cnt = 0;

void define_screen_2_debug(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    static const char* first[] = {"One", "Two", "Three"};
    static int selected_s2d = 0;
    //static const char* second[] = {"Jedna", "Dva", "Tri"};

    INSERT_ELEMENT(new gui_label("Switched to screen 2"))
    INSERT_ELEMENT(new gui_label("Select box test, button (callback) test, graph?",
            -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    EMPTY_LINE

    // elems.push_back(std::unique_ptr<gui_element>(
        // new gui_combobox(second, 3, "Second combobox", -1, -1, 150, app_workspace_ns::font_size::SMALL_FONT)
    // ));
    INSERT_ELEMENT(new gui_combobox(first, 3, &selected_s2d))
    // elems.push_back(std::unique_ptr<gui_element>(
        // new gui_combobox(second, 3, "Second combobox", -1, -1, 200, app_workspace_ns::font_size::BIG_FONT)
    // ));

    EMPTY_LINE
    INSERT_ELEMENT(new gui_button("Test Button", callback_debug, -1, -1, 400, 50))
    if (cb_flag) {
        INSERT_ELEMENT(new gui_label("Press the button!"))
    } else {
        INSERT_ELEMENT(new gui_label("Great! press it again!"))
    }

    static char at[32];
    sprintf(at, "Callback counter: %d", cb_cnt);
    INSERT_ELEMENT(new gui_label(at))
}

void define_screen_3_debug(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    static char fake_in[128];
    static int selected_s3d = 0;

    // following label doesnt work, why?
    INSERT_ELEMENT(new gui_label("Label size TEST small", -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_label("Label size TEST normal", -1, -1, -1, app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_label("Label size TEST big", -1, -1, -1, app_workspace_ns::font_size::BIG_FONT))
    EMPTY_LINE

    INSERT_ELEMENT(new gui_input(IN_TEXT, fake_in, 128, -1, -1, -1, "label", app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_input(IN_TEXT, fake_in, 128, -1, -1, -1, "label", app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_input(IN_TEXT, fake_in, 128, -1, -1, -1, "label", app_workspace_ns::font_size::BIG_FONT))
    EMPTY_LINE

    static const char* combo[] = {"Test", "Pokus"};

    INSERT_ELEMENT(new gui_combobox(combo, 2, &selected_s3d, "Combobox", -1, -1, 150,
                                    app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_combobox(combo, 2, &selected_s3d, "Combobox", -1, -1, 150,
                                    app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_combobox(combo, 2, &selected_s3d, "Combobox", -1, -1, 150,
                                    app_workspace_ns::font_size::BIG_FONT))
    EMPTY_LINE
    INSERT_ELEMENT(new gui_button("TEST", empty_cb, -1, -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_button("TEST", empty_cb, -1, -1, -1, -1, app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_button("TEST", empty_cb, -1, -1, -1, -1, app_workspace_ns::font_size::BIG_FONT))
}

void define_screen_4_debug(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    // elems.push_back(std::unique_ptr<gui_element>(
    //     new gui_direct(render_line)
    // ));
    INSERT_ELEMENT(new gui_label("SMALL", 50, 50, 100, app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_direct(draw_rect))
    INSERT_ELEMENT(new gui_label("Small", 50, 50 + app_workspace_ns::font_size::SMALL_FONT, 100,
                                 app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_direct(draw_rect))
    // elems.push_back(std::unique_ptr<gui_element>(
    //     new gui_direct(render_vert)
    // ));
    INSERT_ELEMENT(new gui_label("NORMAL", 100, 50, 150, app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_direct(draw_rect))
    INSERT_ELEMENT(new gui_label("Normal", 100, 50 + app_workspace_ns::font_size::NORMAL_FONT, 150,
                                 app_workspace_ns::font_size::NORMAL_FONT))
    INSERT_ELEMENT(new gui_direct(draw_rect))
    INSERT_ELEMENT(new gui_label("Large", 220, 50, -1, app_workspace_ns::font_size::BIG_FONT))
    INSERT_ELEMENT(new gui_direct(draw_rect))

    static char in1[32], in2[32], in3[32], in4[32];
    EMPTY_LINE
    INSERT_ELEMENT(new gui_input(IN_TEXT, in1, 32, -1, -1, 50, "in 1"))
    SAME_LINE
    INSERT_ELEMENT(new gui_input(IN_TEXT, in2, 32, -1, -1, 50, "in 2"))
    SAME_LINE
    INSERT_ELEMENT(new gui_input(IN_TEXT, in3, 32, -1, -1, 50, "in 3"))
    INSERT_ELEMENT(new gui_input(IN_TEXT, in4, 32, -1, -1, 50, "in 4"))

    // text wrapping bug fix test
    // INSERT_ELEMENT(new gui_label(get_localized_text("SCREEN_2_SAMPLES_IN_LIMITS"), -1, -1, 100))
    // INSERT_ELEMENT(new gui_direct(text_wrap_debug))
}

void define_screen_5_debug(std::vector<std::unique_ptr<gui_element>>& elems) {
    elems.clear();

    INSERT_ELEMENT(new gui_label("Test/ debug functions dont always have GUI. View logs for information",
                      -1, -1, -1, app_workspace_ns::font_size::SMALL_FONT))
    INSERT_ELEMENT(new gui_button("Test db function", invoke_sql_debug, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("HX samples scale", invoke_sql_debug, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("HX timeout scale", invoke_sql_debug, -1, -1, 460, 40))
    // INSERT_ELEMENT(new gui_label("Scale result: "))
    // SAME_LINE

    INSERT_ELEMENT(new gui_button("Test sql 1", invoke_sql_debug, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test login sql", test_user_login_sql, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test creds sql", test_user_creds_sql, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test headers sql", test_user_meas_header_sql, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test increment sql", test_user_inc_measur_cnt, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test uname ava sql", test_user_is_uname_avail, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Test rfid ava sql", test_user_is_rfid_avail, -1, -1, 460, 40))
    // INSERT_ELEMENT(new gui_button("Err test 1", test_errscr_1, -1, -1, 460, 40))
    // INSERT_ELEMENT(new gui_button("Err test 2", test_errscr_2, -1, -1, 460, 40))

    // INSERT_ELEMENT(new gui_button("Update user rfids", update_rfids_debug, -1, -1, 460, 40))
    // INSERT_ELEMENT(new gui_button("Insert user tst", test_user_insert, -1, -1, 460, 40))
    // INSERT_ELEMENT(new gui_button("Insert measur tst", test_measurement_insert, -1, -1, 460, 40))
    INSERT_ELEMENT(new gui_button("Insert measur tst", test_measurement_query, -1, -1, 460, 40))

}


/*********************************************************************************************************************/
/*       Implementation of local function definitions                                                                */
/*********************************************************************************************************************/

void empty_line() {
    ImGui::NewLine();
}

void same_line() {
    ImGui::SameLine();
}

void begin_scrollabe_region() {
    ImGuiWindowFlags window_flags = ImGuiWindowFlags_HorizontalScrollbar | ImGuiWindowFlags_NavFlattened;
    ImVec2 content_reg = ImGui::GetContentRegionAvail();
    content_reg.y -= 20;
    // ImGui::BeginChild("ScrollableRegion", content_reg, false, window_flags);
    ImGui::BeginChildFrame(ImGui::GetID("ScrollableRegion"), content_reg, window_flags);
}

void end_scrollabe_region() {
    ImGui::EndChildFrame();
}

void login_action() {
    app_workspace::get_instance()->log_in_user_cred();
}

void login_subuser() {
    app_workspace::get_instance()->login_subuser();
}

void logout_action() {
    app_workspace::get_instance()->logoff_user();
}

void logout_subuser() {
    app_workspace::get_instance()->logout_subuser();
}

gui_element* last_element(std::vector<std::unique_ptr<gui_element>> &elems) {
    return elems.at(elems.size() - 1).get();
}

void cb() {
    /*
    app_workspace::get_instance()->user_logged = !app_workspace::get_instance()->user_logged;
    app_workspace::get_instance()->get_scr_mgr()->refresh_screen_elements(1);
    */
   printf("Test callback\n");
}

void draw_rect() {
    ImDrawList *wdl = ImGui::GetWindowDrawList();
    //ImVec2 wpos = ImGui::GetWindowPos();
    ImVec2 lis = ImGui::GetItemRectMin();
    ImVec2 lie = ImGui::GetItemRectMax();
    wdl->AddRect(lis, lie, IM_COL32(255, 0, 0, 255));
    //wdl->AddRect(ImVec2(wpos.x + 200, wpos.y + 300),
    // ImVec2(wpos.x + 350, wpos.y + 300 + (float) app_workspace_ns::font_size::NORMAL_FONT), IM_COL32(255, 0, 0, 255));
}

void int_test_label() {
    ImGui::Text("Int test: %d", app_workspace::get_instance()->int_test);
}

void float_test_label() {
    ImGui::Text("Int test: %f", app_workspace::get_instance()->float_test);
}

void text_test_label() {
    ImGui::Text("Text test: %s", app_workspace::get_instance()->text_test);
}

void render_line() {
    ImDrawList *wdl = ImGui::GetWindowDrawList();
    ImVec2 wpos = ImGui::GetWindowPos();
    wdl->AddLine(ImVec2(wpos.x + 20, wpos.y + 50), ImVec2(wpos.x + 460, wpos.y + 50), IM_COL32(255, 0, 0, 255), 1.0f);
}

void render_vert() {
    ImDrawList *wdl = ImGui::GetWindowDrawList();
    ImVec2 wpos = ImGui::GetWindowPos();
    wdl->AddLine(ImVec2(wpos.x + 98, wpos.y + 50), ImVec2(wpos.x + 98, wpos.y + 74), IM_COL32(0, 0, 0, 255), 1.0f);
}

void empty_cb() {
    printf("empty callback was called\n");
}

void callback_debug() {
    cb_flag = !cb_flag;
    cb_cnt++;
    app_workspace::get_instance()->get_scr_mgr()->refresh_screen_elements(2);
}

void invoke_sql_debug() {
    /*
    app_workspace *app_wrk = app_workspace::get_instance().get();
    uint8_t fake_serial[10] = {0};
    for (int i = 0; i < 5; i++) {
        fake_serial[i] = '1' + i;
    }
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));
    app_wrk->db_conn->query_user_data(usr, nullptr, fake_serial, 5);
    */

//    app_workspace::get_instance()->db_conn->test_insert_with_binary();
   app_workspace::get_instance()->db_conn->test_select_with_binary();
}

void test_user_login_sql() {
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));

    int rv = conn->query_user_data(usr, "test_emp");
    if (rv)
        spdlog::error("screen_definition.cpp - user login test: expected user to be initialized!");
    else
        usr->log_user_to_debug();
}

void test_user_creds_sql() {
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    user_cred creds;

    char passwd_wrong[64] = "This is wrong password";
    char passwd_correct[64] = "9999"; // this password is used on desktop version (only) !!

    rv = conn->query_user_credentials(&creds, "wrong_uname");
    if (rv == 2) {
        spdlog::info("screen_definition.cpp - user cred test: wrong username works as expected.");
    } else {
        spdlog::error("screen_definition.cpp - user cred test: unexpected result for wrong username supplied.");
    }

    rv = conn->query_user_credentials(&creds, "test_emp");
    if (rv) {
        spdlog::error("screen_definition.cpp - user cred test: unexpected failure of correct username");
    } else {
        if (strcmp(creds.passwd, passwd_wrong)) {
            spdlog::info("screen_definition.cpp - user cred test: comparing wrong passwords if false as expected");
        } else {
            spdlog::error("screen_definition.cpp - user cred test: didnt expect true comparison with wrong password");
        }
        if (strcmp(creds.passwd, passwd_correct)) {
            spdlog::error("screen_definition.cpp - user cred test: correct passwd, false comparison");
        } else {
            spdlog::info("screen_definition.cpp - user cred test: correct password, ture comparison");
        }
    }
}

void test_user_meas_header_sql() {
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));

    rv = conn->query_user_data(usr, "test_pac");
    if (rv) {
        spdlog::error("screen_definition.cpp - user header fetch: failed to load user");
        return;
    }
    
    rv = conn->query_measurement_headers(usr);
    if (rv) {
        spdlog::error("screen_definition.cpp - user header fetch: failed to fetch measurement headers");
        return;
    }
    
    for (size_t i = 0; i < usr->measur_headers.size(); i++) {
        measurement_header *h = usr->measur_headers.at(i).get();
        char buf[64] = {0};
        strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(h->measuring_start));
        printf("screen_definition.cpp - user header fetch: header %ld (id: {%ld}), {%s}\n",
                h->measurement_number, h->id, buf);
        // spdlog::info("screen_definition.cpp - user header fetch: header {0} (id: {1}), {3}",
                    //  h->measurement_number, h->id);
    }
}

void test_user_inc_measur_cnt() {
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();

    // 3 = test_adm
    rv = conn->increment_user_measurement_count(3);
    if (rv) {
        spdlog::error("screen_definition.cpp - failed to increment user measurement count");
        return;
    } else {
        spdlog::info("screen_definition.cpp - increment success. Please check db to verify");
    }
}

void test_user_is_uname_avail() {
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();

    rv = conn->is_username_available("test_emp");
    if (rv == 3) {
        spdlog::info("screen_definition.cpp - uname availability test: taken");
    } else {
        spdlog::error("screen_definition.cpp - uname availability test: unexpected rv ({0})", rv);
    }

    rv = conn->is_username_available("asdfjlkzcjv");
    if (rv) {
        spdlog::error("screen_definition.cpp - uname availability test: unexpected rv ({0})", rv);
    } else {
        spdlog::info("screen_definition.cpp - uname availability test: available");
    }
}

void test_user_is_rfid_avail() {
    uint8_t valid[4] = {70, 80, 150, 5};
    uint8_t exist[4] = {0x31, 0x32, 0x33, 0x34};
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();

    rv = conn->is_rfid_serial_available(exist, 4);
    if (rv == 3) {
        spdlog::info("screen_definition.cpp - rfid availability test: taken");
    } else {
        spdlog::error("screen_definition.cpp - rfid availability test: unexpected rv ({0})", rv);
    }

    rv = conn->is_rfid_serial_available(valid, 4);
    if (rv) {
        spdlog::error("screen_definition.cpp - rfid availability test: unexpected rv ({0})", rv);
    } else {
        spdlog::info("screen_definition.cpp - rfid availability test: available");
    }
}

void test_user_insert() {
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));

    usr->role = 2;
    usr->username = "admin";
    usr->name = "a";
    usr->lastname = "b";
    std::time_t time;
    std::time(&time);
    std::tm *dob = std::localtime(&time);
    usr->date_of_birth = *dob;

    int rv = conn->insert_user(usr, "heslo1");
    if (rv) {
        spdlog::error("screen_definition.cpp - user insert test 1: failed");
    }

    usr->role = 0;
    usr->username = "asdf";
    rv = conn->insert_user(usr, "");
    if (rv) {
        spdlog::error("screen_definition.cpp - user insert test 2: failed");
    }
    free(usr);
}

void test_measurement_insert() {
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    measurement *m = (measurement*) calloc(1, sizeof(measurement));

    std::vector<double> v{5.0, 128.782, 0, 1234.1, 123.0};
    m->values = v;
    m->x_val_med = 4.8;
    m->x_val_avg = 5.2;
    m->all_val_med = 8.7;
    m->all_val_avg = 8.5;
    m->measuree_id = 1;
    m->measurer_id = 1;
    m->measurement_number = 5;
    m->measuring_length = 899988123;

    std::time_t time;
    std::time(&time);
    std::tm *a = std::localtime(&time);
    
    m->measuring_start = *a;
    m->measuring_end = *a;

    int rv = conn->insert_measurement(m);
    if (rv) {
        spdlog::error("screen_definition.cpp - measurement insert: failed");
    }

    free(m);
}

void test_measurement_query() {
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    measurement *m = new measurement();

    int rv = conn->query_measurement(m, 6);
    if (rv) {
        spdlog::error("screen_definition.cpp - measurement query: failed");
    }

    delete(m);
}

void update_rfids_debug() {
    int rv;
    db_driver *conn = app_workspace::get_instance()->db_conn.get();
    uint8_t blue_tag[4] = {5, 27, 239, 190}; // read from debug
    uint8_t white_tag[4] = {192, 1, 15, 26}; // read from debug

    rv = conn->update_user_rfid(2, blue_tag, 4); // id = 2 - test employee
    if (rv) {
        spdlog::error("screen_definition.cpp - rfid update failure");
    } else {
        spdlog::info("screen_definition.cpp - rfid update success");
    }

    rv = conn->update_user_rfid(1, white_tag, 4); // id = 1 - test patient
    if (rv) {
        spdlog::error("screen_definition.cpp - rfid update failure");
    } else {
        spdlog::info("screen_definition.cpp - rfid update success");
    }
}

void hx_samples_debug() {

}

void hx_timeout_debug() {

}

void text_wrap_debug() {
    int width = 200;
    int x = -1;
    float wrap_width = (width + (x > -1 ? x : 0)) > 480 ? (480 - x - 10) : width;
    ImVec2 cpos = ImGui::GetCursorPos();
    ImGui::PushTextWrapPos(cpos.x + wrap_width);
    ImGui::Text("22 2 22222222222 22222 222222222222 222222222 22222222222 222 222222");
    ImGui::PopTextWrapPos();
}

void indent() {
    ImGui::Indent(custom_indent);
}

void unindent() {
    ImGui::Unindent(custom_indent);
}

void test_errscr_1() {
    app_workspace::get_instance()->set_err_screen_and_activate(get_localized_text("ERR_USERS_FULL"),
        get_localized_text("ERR_USERS_FULL_DESC"));
    // app_workspace::get_instance()->set_err_screen_and_activate("Error 1",
        // "This is a test. Created by callback 1");
}

void test_errscr_2() {
    app_workspace::get_instance()->set_err_screen_and_activate("Error 2",
        "This is a also test. Created by callback 2");
}

void start_hx_measuring() {
    int rv = app_workspace::get_instance()->hx711_measure();
    if (rv != 0) {
        spdlog::error("screen_definition.cpp - HX measuring callback: measuring failed with error code {0}", rv);
    }
}

void start_hx_continuous() {
    int rv = app_workspace::get_instance()->hx711_continuous_measure();
    if (rv != 0) {
        spdlog::error("screen_definition.cpp - HX measuring callback: failed to start measuring"
            " with error code {0}", rv);
    }
}

void open_measuring_detail() {
    app_workspace *app_wrk = app_workspace::get_instance().get();
    measurement_header* header = 
        app_wrk->userspace->get_measured_user()->measur_headers.at(app_wrk->selected_measuring).get();

    int rv = app_wrk->db_conn->is_measurement_continuous(header->measurement_number, header->measuree_id);
    if (rv == 0) { // continuous measuring
        rv = app_wrk->db_conn->query_continuous_measurement(app_wrk->userspace->continuous_m, 
            header->measurement_number, header->measuree_id);
        if (rv) {
            spdlog::error("screen_definition.cpp - Failed to fetch continuous measurment.");
            app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S4_QUERY_FAILED"), 
                                                 get_localized_text("ERR_S4_SELECTED_MEASUREMENT_CANNOT_DISPLAY"));
            return;
        }

        spdlog::debug("screen_definition.cpp - Successfully retrieved continuous measurement. Preparing to be shown..");

        app_wrk->userspace->picked.reset(new measurement());
        measurement::prepare_continuous_to_picked(app_wrk->userspace->continuous_m, app_wrk->userspace->picked.get());
    } else { // if  measurement is single or error (on error attempt to retrieve single measurement anyway)
        measurement* m = new measurement();
        rv = app_wrk->db_conn->query_measurement(m, header->id);
        if (rv) {
            spdlog::error("screen_definition.cpp - Failed to fetch measurement detail (rv: {0})."
                          " More details from db_driver logs.", rv);
            app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S4_QUERY_FAILED"), 
                                                 get_localized_text("ERR_S4_SELECTED_MEASUREMENT_CANNOT_DISPLAY"));
            return;
        } else {
            spdlog::debug("screen_definition.cpp - Successfully retrieved measurement to be shown");
            app_wrk->userspace->picked.reset(m);
        }
    }

    // fetch username of measuree and measurer and init convienince vars
    std::string uname;
    rv = app_wrk->db_conn->query_username(uname, app_wrk->userspace->picked->measurer_id);
    if (rv) {
        spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measurer username.");
        // failed to retrieve username, but measurement data is already selected
        app_wrk->userspace->picked->init_convinience_vars("DB ERR", "DB ERR");
    } else {
        if (app_wrk->userspace->picked->measurer_id != app_wrk->userspace->picked->measuree_id) {
            std::string sub_uname;
            rv = app_wrk->db_conn->query_username(sub_uname, app_wrk->userspace->picked->measuree_id);

            if (rv) {
                spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measuree username.");
                app_wrk->userspace->picked->init_convinience_vars(uname.c_str(), "DB ERR");
            } else {
                app_wrk->userspace->picked->init_convinience_vars(uname.c_str(), sub_uname.c_str());
            }
        } else {
            app_wrk->userspace->picked->init_convinience_vars(uname.c_str(), uname.c_str());
        }
    }

    app_wrk->get_scr_mgr()->refresh_screen_elements(DETAIL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(DETAIL_SCREEN);
}

void apply_meas_list_filter() {
    app_workspace::get_instance()->get_scr_mgr()->refresh_screen_elements(3);
}

void unit_select_action() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    if (!app_wrk->has_user()) return; // this shouldn't really uccur
    app_wrk->userspace->picked->current_unit = app_wrk->unit_selection[app_wrk->selected_unit_option];
    app_wrk->userspace->picked->init_value_strings();
    app_wrk->refresh_current_screen = true;
}

void move_to_hx_calibration() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    app_wrk->s5_screen_6_indicator = S5_TO_S6_CALIBRATION;
    app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_CTRL_SCREEN);
}

void move_to_user_addition() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    app_wrk->s5_screen_6_indicator = S5_TO_S6_USERADD;
    app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_CTRL_SCREEN);
}

void move_to_hx_test_samp() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    app_wrk->s5_screen_6_indicator = S5_TO_S6_HX_TEST_SAMPLES;
    // app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_CTRL_SCREEN);
    app_wrk->test_hx_samples(); // refresh_screen_elements is called inside this function
}

void move_to_hx_test_time() {
app_workspace *app_wrk = app_workspace::get_instance().get();

    app_wrk->s5_screen_6_indicator = S5_TO_S6_HX_TEST_TIMEOUT;
    // app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_CTRL_SCREEN);
    app_wrk->test_hx_timeout(); // refresh_screen_elements is called inside this function
}

void move_to_kb_delay_test() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    app_wrk->s5_screen_6_indicator = S5_TO_S6_KB_DELAY_TEST;
    app_wrk->set_kb_testing(true);
    app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_CTRL_SCREEN);
}

void start_tare() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    // tare can be ran without having the need for those inputs, but since calibration will follow, check the input
    // here as well 

    if (app_wrk->s6_ref_mass_in <= 0) {
        app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S6_INPUT"),
                                             get_localized_text("ERR_S6_REF_MASS_ERR"));
        return;
    }

    if (app_wrk->s6_samples_in < S6_MIN_SAMPLES_IN || app_wrk->s6_samples_in > S6_MAX_SAMPLES_IN) {
        app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S6_INPUT"),
                                             get_localized_text("ERR_S6_SAMPLES_ERR"));
        return;
    }

    app_wrk->hx711_tare();
}

void start_calibration() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    if (app_wrk->s6_ref_mass_in <= 0) {
        app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S6_INPUT"),
                                             get_localized_text("ERR_S6_REF_MASS_ERR"));
        return;
    }

    if (app_wrk->s6_samples_in < S6_MIN_SAMPLES_IN || app_wrk->s6_samples_in > S6_MAX_SAMPLES_IN) {
        app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S6_INPUT"),
                                             get_localized_text("ERR_S6_SAMPLES_ERR"));
        return;
    }

    app_wrk->s6_ref_mass_str = std::to_string(app_wrk->s6_ref_mass_in) + " " +
                           app_wrk->s6_unit_sel_labels[app_wrk->s6_selected_unit_option];

    app_wrk->hx711_calibrate();
}

void apply_calibration() {
    app_workspace *app_wrk = app_workspace::get_instance().get();

    try {
        app_wrk->hx711_controller->setReferenceUnit(app_wrk->s6_calib_ref_unit);
        app_wrk->hx711_controller->setOffset(app_wrk->s6_calib_offset);
    } catch (std::exception &ex) {
        spdlog::error("screen_definitions.cpp - Failed to apply HX calibration.");
        // FIXME - maybe go to error screen with this
        return;
    }

    config_loader::replace_opt_in_opt_config(app_wrk->main_config->opt_conf_path.c_str(), 
        OPT_CFG_OFFSET_KEY, std::to_string(app_wrk->s6_calib_offset));
    config_loader::replace_opt_in_opt_config(app_wrk->main_config->opt_conf_path.c_str(), 
        OPT_CFG_REF_UNIT_KEY, std::to_string(app_wrk->s6_calib_ref_unit));

    spdlog::info("Successfully applied HX711 calibration");

    app_wrk->s5_success_indicator = S5_CALIBRATION_APPLIED;
    app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_SELECT_SCREEN);
    app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_SELECT_SCREEN);
}

void register_user() {
    app_workspace *app_wrk = app_workspace::get_instance().get();
    if (app_wrk->register_new_user()) {
        spdlog::error("screen_definitions.cpp - Failed to register new user.");
    } else {
        app_wrk->s5_success_indicator = S5_USER_CREATED;
        app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_SELECT_SCREEN);
        app_wrk->get_scr_mgr()->set_selected_screen(ADMIN_SELECT_SCREEN);
    }
}