#include "screen_manager.h"

#include <vector>
#include <exception>
#include <spdlog/spdlog.h>
#include "screen_definitions.h"
#include "imgui.h"
#include "imgui_internal.h" // for access to GetItemID because of focus
#include "localisation.h"
#include "app_workspace.h"

// for debug
// #include "gui_direct.h"

// not including 0 (err_screen): 1-DESIRED_SCREEN_NUMBER
#define DEBUG_SCREENS 5
#define REAL_SCREENS 6

// runtime "window" width (padding not included), and line height (can change for each element)
float rt_content_width;
float rt_line_height;

#define PATIENT_SCREEN_CNT  4
#define EMPLOYEE_SCREEN_CNT 4
// #define ADMIN_SCREEN_CNT    5
const uint8_t patient_screens[] = {1, 2, 3, 4};
const uint8_t employee_screens[] = {1, 2, 3, 4};
// const uint8_t admin_screens[] = {1, 2, 3, 4, 5};

void render_nav_bar();
void render_status_bar();


screen_manager::screen_manager() {
    int desired_screen_number;
    if (app_workspace::get_instance()->is_debug_screens())
        desired_screen_number = DEBUG_SCREENS;
    else
        desired_screen_number = REAL_SCREENS;

    for (int i = 0; i <= desired_screen_number; i++) {
        this->available_screens.emplace(i, std::vector<std::unique_ptr<gui_element>>());
    }

    define_screen_0(available_screens.at(0));
    if (app_workspace::get_instance()->is_debug_screens()) {
        define_screen_1_debug(available_screens.at(1));
        define_screen_2_debug(available_screens.at(2));
        define_screen_3_debug(available_screens.at(3));
        define_screen_4_debug(available_screens.at(4));
        define_screen_5_debug(available_screens.at(5));
    } else {
        define_screen_1(available_screens.at(LOGIN_SCREEN));
        define_screen_2(available_screens.at(MEASURING_SCREEN));
        define_screen_3(available_screens.at(SELECTION_SCREEN));
        define_screen_4(available_screens.at(DETAIL_SCREEN));
        define_screen_5(available_screens.at(ADMIN_SELECT_SCREEN));
        define_screen_6(available_screens.at(ADMIN_CTRL_SCREEN));
    }

    this->selected_screen = 1; // default screen
}

// FIXME : this is duplicate from file event_handler.cpp, but is only used for debug here
template<typename Base, typename T>
inline bool instanceof(const T *ptr) {
    return dynamic_cast<const Base*>(ptr) != nullptr;
}

void screen_manager::render_screen(std::vector<std::unique_ptr<gui_element>> &elements) {
    focused = nullptr;
    activated = nullptr;
    ImGuiID last_focused = 0;

    for (size_t i = 0; i < elements.size(); i++) {
        std::unique_ptr<gui_element> &element = elements.at(i);
        element->index = i;
        element->render_element();

        if (ImGui::IsItemFocused()) {
            // maybe GetFocusID instead?
            if (last_focused != ImGui::GetItemID()) {
                focused = element.get();
                last_focused = ImGui::GetItemID();
            }
        }
        /*// debug if
        if (ImGui::IsItemFocused()) {
            if (instanceof<gui_direct>(element.get())) {
                ImGui::IsItemFocused();
                printf("Wrong item is focused\n");
            }
            // if (i == 7) { // this should be gui_direct and cannot be higlighted
                // ImGui::IsItemFocused();
            // }
            // if (i == 6) { // this should be password input
                // ImGui::IsItemFocused();
            // }
            spdlog::debug("setting focus item index {0}", i);
            focused = element.get();
        }
        */
        // activated is only true when the item was previously inactive
        else if (ImGui::IsItemActivated() || ImGui::IsItemActive()) 
            activated = element.get();
    }
}

bool screen_manager::is_screen_available_to_user(const uint8_t *screens, const uint8_t count, const uint8_t requested) {
    for (int i = 0; i < count; i++) {
        if (screens[i] == requested)
            return true;
    }
    return false;
}

void screen_manager::render_active_screen() {
    //spdlog::debug("screen_manager.cpp - Rendering screen number: {0}", this->selected_screen);
    rt_content_width = ImGui::GetWindowContentRegionMax().x;
    rt_line_height = ImGui::GetTextLineHeightWithSpacing();

    if (app_workspace::get_instance()->refresh_current_screen) {
        refresh_screen_elements(selected_screen);
        app_workspace::get_instance()->refresh_current_screen = false;
    }

    /*
    if (ImGui::IsAnyItemActive())
        spdlog::info("An item is ACTIVE");
    if (ImGui::IsAnyItemFocused())
        spdlog::info("An item is FOCUSED");
    if (ImGui::IsAnyItemHovered())
        spdlog::info("An item is HOVERED");
    */

    app_workspace::get_instance()->use_font_size(app_workspace_ns::font_size::SMALL_FONT);
    render_status_bar();

    app_workspace::get_instance()->use_font_size(app_workspace_ns::font_size::NORMAL_FONT);
    try {
        std::vector<std::unique_ptr<gui_element>> &elements = available_screens.at(selected_screen);
        render_screen(elements);
    } catch (std::out_of_range &ex) {
        spdlog::error("screen_manager.cpp - Failed to find screen key in available_screens map."
                      " Cannot display selected screen");
        selected_screen = 0;
        render_active_screen();
        return;
    }

    app_workspace::get_instance()->use_font_size(app_workspace_ns::font_size::SMALL_FONT);
    render_nav_bar();
}

uint8_t screen_manager::get_selected_screen() {
    return this->selected_screen;
}

void screen_manager::set_selected_screen(uint8_t screen) {
    /* not tested - for now disabled for easier testing
    if (app_workspace::get_instance()->has_user()) {
        if (app_workspace::get_instance()->userspace->get_role() == 0) {
            if (!is_screen_available_to_user(patient_screens, PATIENT_SCREEN_CNT, screen)) {
                spdlog::debug("User doesn't have permission to access screen {0}", screen);
                return;
            }
        } else if (app_workspace::get_instance()->userspace->get_role() == 1) {
            if (!is_screen_available_to_user(employee_screens, EMPLOYEE_SCREEN_CNT, screen)) {
                spdlog::debug("User doesn't have permission to access screen {0}", screen);
                return;
            }
        } else if (app_workspace::get_instance()->userspace->get_role() == 2) {
            // no action for admin, admin has permission for all screens
        } else {
            spdlog::error("User role not recognized when setting screen");
            return;
        }
    } else {
        spdlog::debug("No user login, only available screen is 1");
        if (this->selected_screen != 1) this->selected_screen = 1;
        return;
    }
    */

    if (app_workspace::get_instance()->is_debug_screens()) {
        if (screen > DEBUG_SCREENS) {
            spdlog::debug("Discarding screen select request.");
            return;
        }
    } else {
        if (screen > REAL_SCREENS) {
            spdlog::debug("Discarding screen select request.");
            return;
        }
    }

    spdlog::info("Setting screen to {0}", screen);
    push_screen_to_previous();
    this->selected_screen = screen;
}


void screen_manager::refresh_screen_elements(uint8_t screen) {
    std::vector<std::unique_ptr<gui_element>> *elems;

    try {
        elems = &(available_screens.at(screen));
        //elems->clear(); //clear current screen elements
    } catch (std::invalid_argument &ex) {
        spdlog::error("screen_manager.cpp - Cannot refresh screen {0}. Screen doesn't exist");
        return;
    }

    if (app_workspace::get_instance()->is_debug_screens()) {
        switch (screen) {
            case 0:
                define_screen_0(*elems); break;
            case 1:
                define_screen_1_debug(*elems); break;
            case 2:
                define_screen_2_debug(*elems); break;
            case 3:
                define_screen_3_debug(*elems); break;
            case 4:
                define_screen_4_debug(*elems); break;
            case 5:
                define_screen_5_debug(*elems); break;

            default:
                spdlog::error("screen_manager.cpp - Cannot refresh screen {0}. Screen doesn't exist", screen);
                break;
        }
    } else {
        switch (screen) {
            case 0:
                define_screen_0(*elems); break;
            case 1:
                define_screen_1(*elems); break;
            case 2:
                define_screen_2(*elems); break;
            case 3:
                define_screen_3(*elems); break;
            case 4:
                define_screen_4(*elems); break;
            case 5:
                define_screen_5(*elems); break;
            case 6:
                define_screen_6(*elems); break;

            default:
                spdlog::error("screen_manager.cpp - Cannot refresh screen {0}. Screen doesn't exist", screen);
                break;
        }
    }
}

void screen_manager::refresh_all_screns() {
    // starting from 1, because 0 (err_sreen) doesn't need to be reloaded
    // also don't do error checking for i > 255 (refresh_screen_elements takes uint8_t)
    for (size_t i = 1; i < available_screens.size(); i++) {
        refresh_screen_elements(i);
    }
}

void screen_manager::push_screen_to_previous() {
    if (prev_screen_stack.size() > PREV_SCREEN_Q_MAX)
        prev_screen_stack.erase(prev_screen_stack.begin() + (PREV_SCREEN_Q_MAX / 2));
        // prev_screen_stack.clear();
    
    // dont allow to go back to error screen
    if (selected_screen > 0)
        prev_screen_stack.push_back(selected_screen);
}

void screen_manager::pop_screen_from_previous() {
    if (prev_screen_stack.size() <= 0) {
        spdlog::info("screen_manager.cpp - Cannot go to previous screen. No screens in stack.");
        return;
    }

    selected_screen = prev_screen_stack[prev_screen_stack.size() - 1];
    prev_screen_stack.pop_back();
}

void screen_manager::clear_previous_screen_stack() {
    prev_screen_stack.clear();
}

/*********************************************************************************************************************/
/*      private screen_manager functions                                                                             */
/*********************************************************************************************************************/

///////////////////////////////////////////////////////////////////////

void draw_status_circle(ImVec2 center, bool usr_logged, float radius) {
    ImU32 color = usr_logged ? IM_COL32(0, 255, 0, 255) : IM_COL32(255, 0, 0, 255);

    ImDrawList *wdl = ImGui::GetWindowDrawList();
    wdl->AddCircleFilled(center, radius, color);
}

void render_status_bar() {
    app_workspace *app_wrk = app_workspace::get_instance().get();
    screen_manager *scr_mgr = app_wrk->get_scr_mgr();
    ImVec2 win_pos = ImGui::GetWindowPos();
    ImVec2 win_size = ImGui::GetContentRegionMax();
    ImVec2 cur_pos;
    ImVec2 padding = ImGui::GetStyle().WindowPadding;
    char usr_buff[DEF_BUFF_SIZE] = {0};
    float radius = 8;

    sprintf(usr_buff, "%s: %s", get_localized_text("GT_USER"), 
                                app_wrk->has_user() ? app_wrk->userspace->get_username().c_str() :
                                                      get_localized_text("GT_USER_NO_LOGIN"));

    cur_pos = ImGui::GetCursorPos();
    ImVec2 usr_center(win_pos.x + cur_pos.x + radius, win_pos.y + cur_pos.y + (radius / 2));

    // print username and draw status circle
    draw_status_circle(usr_center, app_wrk->user_logged, radius);
    ImGui::SetCursorPosX(padding.x + (radius * 2) + padding.x);
    ImGui::Text(usr_buff);

    // print time
    time_t now = time(0);
    tm *ltm = localtime(&now);
    char time[DEF_BUFF_SIZE_V_SMALL];
    sprintf(time, "%.2d:%.2d:%.2d", ltm->tm_hour, ltm->tm_min, ltm->tm_sec);
    ImVec2 tts = ImGui::CalcTextSize(time); // time text size
    ImGui::SameLine();
    float time_x = win_size.x - padding.x - tts.x - 2;
    ImGui::SetCursorPosX(time_x);
    ImGui::Text(time);

    // print selected screen
    // FIXME centered in previous version, moved more to right because i haven't decided yet on  usr: usr_name length
    ImGui::SameLine();
    //ImGui::SetCursorPosX((win_size.x / 2) - padding.x - 2);
    ImGui::SetCursorPosX(time_x - 100);
    ImGui::TextColored(ImColor(IM_COL32(0, 180, 0, 255)), "%d", scr_mgr->get_selected_screen());

    // "extra" empty line after status bar (as padding)
    ImGui::NewLine();
}

/**
 * @brief called nav bar, render current input mode at the bottom of the screen
 * 
 */
void render_nav_bar() {
    app_workspace *app_wrk = app_workspace::get_instance().get();;
    app_workspace_ns::kb_input_state *kbins = app_wrk->kb_in_state.get();
    ImVec2 win_size = ImGui::GetContentRegionMax();
    char buff[DEF_BUFF_SIZE_BIG] = {0}, in_mod[DEF_BUFF_SIZE_SMALL] = {0};
    float th = ImGui::GetTextLineHeightWithSpacing(); // text height (with spacing)

    switch (kbins->current_mode) {
        case app_workspace_ns::kb_input_mode::NAVIGATION:
            strcpy(in_mod, get_localized_text("GT_NAVIGATION"));
            break;
        case app_workspace_ns::kb_input_mode::INPUT:
            switch (kbins->current_type) {
                case app_workspace_ns::kb_input_type::INT:
                    strcpy(in_mod, get_localized_text("GT_INPUT_INT"));
                    break;
                case app_workspace_ns::kb_input_type::FLOAT:
                    strcpy(in_mod, get_localized_text("GT_INPUT_FLOAT"));
                    break;
                case app_workspace_ns::kb_input_type::TEXT:
                    strcpy(in_mod, get_localized_text("GT_INPUT_TEXT"));
                    break;
                
                case app_workspace_ns::kb_input_type::NONE:
                default:
                    strcpy(in_mod, get_localized_text("GT_INPUT_ERR"));
                    break;
            }
            break;
        default:
            strcpy(in_mod, get_localized_text("GT_ERR"));
            break;
    }

    sprintf(buff, "%s: %s\t%s: %s\t%s: %s", 
        get_localized_text("GT_MODE"),
        in_mod, 
        get_localized_text("GT_CAPS"),
        app_wrk->get_instance()->capslock_flag ? get_localized_text("GT_YES") : get_localized_text("GT_NO"),
        get_localized_text("GT_SHIFT"),
        app_wrk->get_instance()->lshift_flag   ? get_localized_text("GT_YES") : get_localized_text("GT_NO"));

    ImGui::SetCursorPosY(win_size.y - th);
    ImGui::Text(buff);
}


uint8_t screen_manager::get_screen_count() {
    // -1 because 0 is err_screen
    return available_screens.size() - 1;
}