#include "custom_hx711.h"

#include <thread>

using namespace HX711;


custom_watcher::custom_watcher(HX711::HX711 *hx) : Watcher(hx) {}

WatchState custom_watcher::get_watch_state() {
    return this->_watchState;
}

custom_hx711::custom_hx711(
    const int dataPin,
    const int clockPin,
    const Value refUnit,
    const Value offset,
    const Rate rate) : 
        AdvancedHX711(dataPin, clockPin, refUnit, offset, rate) 
{
    this->_wx = new custom_watcher(this);
    this->_wx->begin();
}

void custom_hx711::stop_reading() {
    this->_wx->pause();
}

std::vector<Value> custom_hx711::getValues(const std::chrono::nanoseconds timeout) {

    using namespace std::chrono;

    this->_wx->values.clear();
    this->_wx->watch();

    std::vector<Value> vals;
    const auto endTime = steady_clock::now() + timeout;

    while(true) {

        if(steady_clock::now() >= endTime) {
            this->_wx->pause();
            return vals;
        }

        if (this->_wx->get_watch_state() == WatchState::PAUSE) {
            return vals;
        }

        if(!this->_wx->values.empty()) {

            this->_wx->valuesLock.lock();
            while(!this->_wx->values.empty()) {
                vals.push_back(this->_wx->values.pop());
            }
            this->_wx->valuesLock.unlock();

            continue;

        }

        std::this_thread::yield();
        Utility::sleep(milliseconds(1));

    }

}

std::vector<Value> custom_hx711::getValues(const std::size_t samples) {

    using namespace std::chrono;

    if(samples == 0) {
        throw std::range_error("samples must be at least 1");
    }

    this->_wx->values.clear();
    this->_wx->watch();

    std::vector<Value> vals;
    vals.reserve(samples);

    //while not filled
    while(vals.size() < samples) {

        //while empty, defer exec to other threads
        while(this->_wx->values.empty() && this->_wx->get_watch_state() != WatchState::PAUSE) {
            std::this_thread::yield();
            Utility::sleep(milliseconds(1));
        }

        if (this->_wx->get_watch_state() == WatchState::PAUSE) {
            return vals;
        }

        //not empty; data available!
        this->_wx->valuesLock.lock();

        //now, take as many values as which are available
        //up to however many are left to fill the array
        while(!this->_wx->values.empty() && vals.size() < samples) {
            vals.push_back(this->_wx->values.pop());
        }

        this->_wx->valuesLock.unlock();

    }

    this->_wx->pause();

    return vals;

}

std::vector<Value> custom_hx711::getValues(const std::chrono::nanoseconds timeout, std::vector<Value> **watchdog) {
    using namespace std::chrono;

    this->_wx->values.clear();
    this->_wx->watch();

    std::vector<Value> vals;
    *watchdog = &vals;
    const auto endTime = steady_clock::now() + timeout;

    while(true) {

        if(steady_clock::now() >= endTime) {
            this->_wx->pause();
            return vals;
        }

        if (this->_wx->get_watch_state() == WatchState::PAUSE) {
            return vals;
        }

        if(!this->_wx->values.empty()) {

            this->_wx->valuesLock.lock();
            while(!this->_wx->values.empty()) {
                vals.push_back(this->_wx->values.pop());
            }
            this->_wx->valuesLock.unlock();

            continue;

        }

        std::this_thread::yield();
        Utility::sleep(milliseconds(1));

    }
}

std::vector<Value> custom_hx711::getValues(const std::size_t samples, std::vector<Value> **watchdog) {
    using namespace std::chrono;

    if(samples == 0) {
        throw std::range_error("samples must be at least 1");
    }

    this->_wx->values.clear();
    this->_wx->watch();

    std::vector<Value> vals;
    *watchdog = &vals;
    vals.reserve(samples);

    //while not filled
    while(vals.size() < samples) {

        //while empty, defer exec to other threads
        while(this->_wx->values.empty() && this->_wx->get_watch_state() != WatchState::PAUSE) {
            std::this_thread::yield();
            Utility::sleep(milliseconds(1));
        }

        if (this->_wx->get_watch_state() == WatchState::PAUSE) {
            return vals;
        }

        //not empty; data available!
        this->_wx->valuesLock.lock();

        //now, take as many values as which are available
        //up to however many are left to fill the array
        while(!this->_wx->values.empty() && vals.size() < samples) {
            vals.push_back(this->_wx->values.pop());
        }

        this->_wx->valuesLock.unlock();

    }

    this->_wx->pause();

    return vals;
}