#include <spdlog/spdlog.h>
#include "event_handler.h"
#include "app_workspace.h"
#include "gui_input.h"
#include "gui_button.h"
#include "gui_combobox.h"
#include "localisation.h"
#include "imgui.h"
#include "imgui_internal.h"

#include "gui_direct.h"
#include "gui_label.h"
#include "gui_observer.h"
#include "gui_selectable.h"

namespace event_handler {

template<typename Base, typename T>
inline bool instanceof(const T *ptr) {
    return dynamic_cast<const Base*>(ptr) != nullptr;
}

/*********************************************************************************************************************/
/*       Global variables                                                                                            */
/*********************************************************************************************************************/

    /*
    bool lshift_flag = false;
    bool capslock_flag = false;
    */

    // SPI keyboard representation (its actually keypad)
    // # (0x2000) is ONLY USED as "enter" confirm/ select button
    /******************************************************************************/
    /*                                                                            */
    /*    / ---------- \    / ---------- \    / ---------- \    / ---------- \    */
    /*    | 1 = 0x0008 |    | 2 = 0x0004 |    | 3 = 0x0002 |    | A = 0x0001 |    */
    /*    \ ---------- /    \ ---------- /    \ ---------- /    \ ---------- /    */
    /*                                                                            */
    /*    / ---------- \    / ---------- \    / ---------- \    / ---------- \    */
    /*    | 4 = 0x0080 |    | 5 = 0x0040 |    | 6 = 0x0020 |    | B = 0x0010 |    */
    /*    \ ---------- /    \ ---------- /    \ ---------- /    \ ---------- /    */
    /*                                                                            */
    /*    / ---------- \    / ---------- \    / ---------- \    / ---------- \    */
    /*    | 7 = 0x0800 |    | 8 = 0x0400 |    | 9 = 0x0200 |    | C = 0x0100 |    */
    /*    \ ---------- /    \ ---------- /    \ ---------- /    \ ---------- /    */
    /*                                                                            */
    /*    / ---------- \    / ---------- \    / ---------- \    / ---------- \    */
    /*    | * = 0x8000 |    | 0 = 0x4000 |    | # = 0x2000 |    | D = 0x1000 |    */
    /*    \ ---------- /    \ ---------- /    \ ---------- /    \ ---------- /    */
    /*                                                                            */
    /******************************************************************************/

    /**
     * @brief Map for translating SPI keyboard scancodes keys to SDL scancodes for navigation.
     * 
     */
    std::map<uint16_t, SDL_Scancode> navigation_keymap {
        {0x0008, SDL_SCANCODE_1},
        {0x0004, SDL_SCANCODE_2},
        {0x0002, SDL_SCANCODE_3},
        {0x0001, SDL_SCANCODE_UP}, // use A as UP arrow
        //        
        {0x0080, SDL_SCANCODE_4},
        {0x0040, SDL_SCANCODE_5},
        {0x0020, SDL_SCANCODE_6},
        {0x0010, SDL_SCANCODE_LEFT}, // use B as LEFT arrow (possibly bad contact, doesn't always work)
        //        
        {0x0800, SDL_SCANCODE_7},
        {0x0400, SDL_SCANCODE_8},
        {0x0200, SDL_SCANCODE_9},
        {0x0100, SDL_SCANCODE_RIGHT}, // use C as RIGHT arrow (possibly bad contact, doesn't always work)
        //        
        //{0x8000, SDL_SCANCODE_KP_MULTIPLY},
        {0x4000, SDL_SCANCODE_BACKSPACE},
        //{0x2000, SDL_SCANCODE_RETURN}, // use # as RETURN - needs to be same for all maps, maybe take out?
        {0x1000, SDL_SCANCODE_DOWN} // use D as DOWN arrow
    };

    /**
     * @brief int input map that is used for mapping spi keys to sdl scancodes
     * 
     */
    std::map<uint16_t, SDL_Scancode> int_in_keymap {
        {0x0008, SDL_SCANCODE_1},
        {0x0004, SDL_SCANCODE_2},
        {0x0002, SDL_SCANCODE_3},
        {0x0001, SDL_SCANCODE_BACKSPACE}, // use A as BACKSPACE (wanted this on B, but B isn't working properly)
        //        
        {0x0080, SDL_SCANCODE_4},
        {0x0040, SDL_SCANCODE_5},
        {0x0020, SDL_SCANCODE_6},
        {0x0010, SDL_SCANCODE_MINUS},
        //        
        {0x0800, SDL_SCANCODE_7},
        {0x0400, SDL_SCANCODE_8},
        {0x0200, SDL_SCANCODE_9},
        //{0x0100, SDL_SCANCODE_C},
        //        
        //{0x8000, SDL_SCANCODE_KP_MULTIPLY},
        {0x4000, SDL_SCANCODE_0},
        //{0x2000, SDL_SCANCODE_RETURN}, // use # as RETURN - needs to be same for all maps, maybe take out?
        //{0x1000, SDL_SCANCODE_D}
    };

    /**
     * @brief float input map that is used for mapping spi keys to sdl scancodes
     * 
     */
    std::map<uint16_t, SDL_Scancode> float_in_keymap {
        {0x0008, SDL_SCANCODE_1},
        {0x0004, SDL_SCANCODE_2},
        {0x0002, SDL_SCANCODE_3},
        {0x0001, SDL_SCANCODE_BACKSPACE}, // use A as BACKSPACE (wanted this on B, but B isn't working properly)
        //        
        {0x0080, SDL_SCANCODE_4},
        {0x0040, SDL_SCANCODE_5},
        {0x0020, SDL_SCANCODE_6},
        {0x0010, SDL_SCANCODE_MINUS},
        //        
        {0x0800, SDL_SCANCODE_7},
        {0x0400, SDL_SCANCODE_8},
        {0x0200, SDL_SCANCODE_9},
        //{0x0100, SDL_SCANCODE_C},
        //        
        {0x8000, SDL_SCANCODE_DECIMALSEPARATOR}, // uses * as decimal separator
        {0x4000, SDL_SCANCODE_0},
        //{0x2000, SDL_SCANCODE_RETURN}, // use # as RETURN - needs to be same for all maps, maybe take out?
        //{0x1000, SDL_SCANCODE_D}
    };

    /**
     * @brief Text input keymap. Holds what scancodes are cycled for corresponding key.
     * First value should correspond to the physical key value ("picture" on keyboard).
     * Some keys are currently disabled. Either because they are reserved or have no use.
     * 
     */
    std::map<uint16_t, std::vector<SDL_Scancode>> text_in_keymap {
        {0x0008, {SDL_SCANCODE_1}},
        {0x0004, {SDL_SCANCODE_2, SDL_SCANCODE_A, SDL_SCANCODE_B, SDL_SCANCODE_C}},
        {0x0002, {SDL_SCANCODE_3, SDL_SCANCODE_D, SDL_SCANCODE_E, SDL_SCANCODE_F}},
        {0x0001, {SDL_SCANCODE_BACKSPACE}}, // same as above
        //        
        {0x0080, {SDL_SCANCODE_4, SDL_SCANCODE_G, SDL_SCANCODE_H, SDL_SCANCODE_I}},
        {0x0040, {SDL_SCANCODE_5, SDL_SCANCODE_J, SDL_SCANCODE_K, SDL_SCANCODE_L}},
        {0x0020, {SDL_SCANCODE_6, SDL_SCANCODE_M, SDL_SCANCODE_N, SDL_SCANCODE_O}},
        {0x0010, {SDL_SCANCODE_LSHIFT}},
        //        
        {0x0800, {SDL_SCANCODE_7, SDL_SCANCODE_P, SDL_SCANCODE_Q, SDL_SCANCODE_R, SDL_SCANCODE_S}},
        {0x0400, {SDL_SCANCODE_8, SDL_SCANCODE_T, SDL_SCANCODE_U, SDL_SCANCODE_V}},
        {0x0200, {SDL_SCANCODE_9, SDL_SCANCODE_W, SDL_SCANCODE_X, SDL_SCANCODE_Y, SDL_SCANCODE_Z}},
        {0x0100, {SDL_SCANCODE_CAPSLOCK}},
        //        
        {0x8000, {SDL_SCANCODE_KP_MULTIPLY, SDL_SCANCODE_DECIMALSEPARATOR, SDL_SCANCODE_MINUS}},
        {0x4000, {SDL_SCANCODE_0}},
        //{0x2000, {SDL_SCANCODE_KP_HASH}}, // MUSTNT USER THIS!!!
        //{0x1000, {SDL_SCANCODE_D}}
    };

    /** This map is used to create a character input for SDL scancode */
    std::map<SDL_Scancode, const char*> custom_keymap {
        {SDL_SCANCODE_KP_MULTIPLY, "*"},
        {SDL_SCANCODE_DECIMALSEPARATOR, "."},
        {SDL_SCANCODE_MINUS, "-"}
    };

    /** This map is used when the SHIFT keymoad is toggled (true) */
    std::map<SDL_Scancode, const char*> custom_shift_keymap {
        {SDL_SCANCODE_MINUS, "_"}
    };

/*********************************************************************************************************************/
/*       Local function definitions                                                                                  */
/*********************************************************************************************************************/

    /**
     * @brief Determines SDL_Scancode from SPI keyboard keycode and cycle. Cycle is used
     * to cycle keys during text input.
     * 
     * @param keycode SPI keyboard 
     * @param cycle repeted pressed withnin a time limit
     * @return SDL_Scancode 
     */
    SDL_Scancode determine_scancode(uint16_t keycode, uint8_t cycle = 0);
    /**
     * @brief Creates an SDL_Event simulating keyboard press from SDL_Scancode
     * 
     * @param scancode scancode of keyboard key
     * @param mod 
     * @return SDL_Event 
     */
    SDL_Event create_key_event(SDL_Scancode scancode, Uint16 mod = 0);
    /**
     * @brief Creates an SDL_Event for input
     * 
     * @param scancode input value
     * @return SDL_Event 
     */
    SDL_Event create_textinput_event(SDL_Scancode scancode);
    /**
     * @brief Pushes event to SDL queue. This event then can be propagated to ImGui. If @replace_cur_char is set,
     * ten a BACKSPACE event is pushed first.
     * 
     * @param event 
     * @param replace_cur_char 
     */
    void push_event_to_sdl(SDL_Event* event, bool replace_cur_char = false);

    /**
     * @brief Determine if keyboard can be switched to INPUT mode
     */
    bool can_switch_modes();
    /**
     * @brief Handles return event (NAVIGATION/INPUT mode switch, BUTTON/ COMBOOX... activation)
     */
    int handle_return_event();
    /**
     * @brief Handle event as navigation event
     */
    int handle_nav_event(kb_event* raw_event);
    /**
     * @brief Handle event as input event.
     */
    int handle_input_event(kb_event* raw_event);

/*********************************************************************************************************************/
/*       Implementation of header function definitions                                                               */
/*********************************************************************************************************************/

    // FIXME - move those to Global variables section
    static kb_event* last_raw_event = nullptr;
    static uint8_t key_cycle = 0;
    // this flag is needed because ImGui can't tell, that an combobox is focused (expanded)
    static bool combo_expanded = false;

    /**
     * @brief Offers option for cutom event handling. If not desired only created SDL event.
     * 
     * @param raw_event SPI keyboard event
     * @return int - currently not used anywhere
     */
    int handle_raw_event(kb_event* raw_event) {
        if (app_workspace::get_instance()->hx_measuring || app_workspace::get_instance()->get_kb_testing()) {
            if (raw_event->scancode == 0x2000 &&
                !(raw_event->flags & KB_FLAGS_REPEATING) && (raw_event->flags & KB_FLAGS_KEY_DOWN))
            {
                if (app_workspace::get_instance()->hx_measuring) {
                    app_workspace::get_instance()->hx711_controller->stop_reading();
                    app_workspace::get_instance()->hx_continuous = false;
                    app_workspace::get_instance()->hx_not_finished_yet = true;
                    app_workspace::get_instance()->refresh_current_screen = true;
                    spdlog::info("event_handler.cpp - Recieved event to stop continuous measuring. Stopping...");
                } else {
                    app_workspace::get_instance()->set_kb_testing(false);
                    app_workspace::get_instance()->get_scr_mgr()->set_selected_screen(ADMIN_SELECT_SCREEN);
                    app_workspace::get_instance()->get_scr_mgr()->refresh_screen_elements(ADMIN_SELECT_SCREEN);
                }
                return 0;
            }

            spdlog::debug("event_handler.cpp - Not handling keyboard events because a hx measuring is ongoing.");
            return 0;
        }

        if (raw_event->flags & KB_FLAGS_TEST) {
            return 0;
        }

        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;

        spdlog::debug("Handling raw event: {0}", raw_event->scancode);

        // handling '#' used as return - always the same action (considered key exception, doesn't go to lookup maps)
        // only handle on first event KEY DOWN
        if (raw_event->scancode == 0x2000 && 
            !(raw_event->flags & KB_FLAGS_REPEATING) && 
            (raw_event->flags & KB_FLAGS_KEY_DOWN))
        {
            return handle_return_event();
        }

        if (raw_event->flags & KB_FLAGS_KEY_DOWN) { // handling KEY_DOWN events
            if (!(raw_event->flags & KB_FLAGS_REPEATING)) { // only first event (not continous hold events)
                if (kbins->current_mode == app_workspace_ns::kb_input_mode::NAVIGATION)
                    return handle_nav_event(raw_event);
                    
                if (kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT)
                    return handle_input_event(raw_event);
            }
        }

        return 1;
    }
    
    /*
        Return 0 if only custom handling of the event is enough. 1 if the event is to be forwarded to ImGui.
    */
    int handle_sdl_event(const SDL_Event* event) {
        screen_manager *scr_mgr = app_workspace::get_instance()->get_scr_mgr();
        // reference to KeyBoardINputState (kbins)
        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;

        // if no item is focused, arrow down focuses "second" element and arrow up focuses "first" element
        // this masks arrow down to arrow up in case when no item is focused resulting in first element being focused
        // FIXME - potential bug: arrow down KEYUP event is still processed, this triggers on KEYDOWN and 
        // doesn't trigger again on KEYUP - event seq: UP - KEYDOWN, UP - KEYUP, DOWN - KEYUP
        if ((event->type == SDL_KEYDOWN || event->type == SDL_KEYUP) && 
            (event->key.keysym.scancode == SDL_SCANCODE_DOWN || event->key.keysym.scancode == SDL_SCANCODE_UP) &&
            (!ImGui::IsAnyItemFocused() || (!scr_mgr->focused && !combo_expanded)))
        {
            // need to consume event and reque new one, only changing current event doesn't work
            spdlog::debug("event_handler.cpp - masking arrow key DOWN as UP {0}", combo_expanded);
            //SDL_Event e = create_key_event(SDL_SCANCODE_UP);
            // "fix" - after changing screens focus is lost, but focused item count is remembered and incorrect item is selected
            // PAGE_UP selects first item
            SDL_Event e = create_key_event(SDL_SCANCODE_PAGEUP);
            push_event_to_sdl(&e, false);

            return 0;
        }

        // handling only KEYDOWN events
        if (event->type == SDL_KEYDOWN) {
            spdlog::debug("event_handler.cpp - KEYDOWN scancode: {0}", event->key.keysym.scancode);

// raspbian should be defined on rpi through cmake option
#ifndef __RASPBIAN__
            if (event->key.keysym.scancode == SDL_SCANCODE_RALT) {
                handle_return_event();
                return 0;
            }

            if (kbins->current_mode == app_workspace_ns::kb_input_mode::NAVIGATION && 
                event->key.keysym.scancode == SDL_SCANCODE_BACKSPACE)
            {
                scr_mgr->pop_screen_from_previous();
                return 0;
            }
#endif

            // handle 1-9 while kb_input_mode is NAVIGATION (0 isn't included)
            // need this for navigation on desktop, but for rpi moved to handle_nav_event() and never comes here
            if (kbins->current_mode == app_workspace_ns::kb_input_mode::NAVIGATION &&
                event->key.keysym.scancode >= SDL_SCANCODE_1 &&
                event->key.keysym.scancode < SDL_SCANCODE_0) {
                // screen number is only allowed in range 1-9 (0 reserved for error)
                uint8_t screen_no = strtol(SDL_GetKeyName(event->key.keysym.sym), NULL, 10);

                if (screen_no > scr_mgr->get_screen_count()) {
                    spdlog::debug("event_handler.cpp - Requested screen ({0}) doesn't exist", screen_no);
                    return 0;
                }

                scr_mgr->set_selected_screen(screen_no);

                return 0;
            }

            // handle other keys (than 1-9) while in NAVIGATION mode
            if (kbins->current_mode == app_workspace_ns::kb_input_mode::NAVIGATION) {
                return 1;
            }

            // handle any keypress (exc. #) while kb_input_mode is INPUT
            if (kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT) {
            }
        }

        if (event->type == SDL_KEYUP) {
            spdlog::debug("event_handler.cpp - KEYUP scancode: {0}", event->key.keysym.scancode);
            return 1; // DO NOT SEND KEYUP TO IMGUI - this is to prevent unwanted actions on desktop (spi keyboard only sends keydown events)
        }

        return 1;
    }

/*********************************************************************************************************************/
/*       Implementation of local function definitions                                                                */
/*********************************************************************************************************************/

    SDL_Scancode determine_scancode(uint16_t keycode, uint8_t cycle) {
        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;

        try {
            if (kbins->current_mode == app_workspace_ns::kb_input_mode::NAVIGATION) {
                return navigation_keymap.at(keycode);
            } else if (kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT) {
                switch (kbins->current_type) {
                    case app_workspace_ns::kb_input_type::FLOAT: {
                            return float_in_keymap.at(keycode);
                        } break;
                    case app_workspace_ns::kb_input_type::INT: {
                            return int_in_keymap.at(keycode);
                        } break;
                    case app_workspace_ns::kb_input_type::TEXT: {
                            std::vector<SDL_Scancode> opts = text_in_keymap.at(keycode);
                            return opts[cycle % opts.size()];
                        } break;
                    case app_workspace_ns::kb_input_type::NONE:
                    default: 
                        spdlog::error("keyboard.cpp - This kb_input_type has no keymap");
                }
            } else {
                spdlog::error("keyboard.cpp - Unknown kb_input_mode");
            }
        } catch (const std::out_of_range& ex) {
            /* logging disabled, because of many key options, that are not specified in maps
            spdlog::error("keyboard.cpp - Failed to find event for pressed key.");
            spdlog::error("\t\t\t - 1) Key doesn't have event in selected input mode and type");
            spdlog::error("\t\t\t - 2) Multiple keys pressed at once (unrecognized key value)");
            */
            return SDL_SCANCODE_CLEAR;
        }

        return SDL_SCANCODE_CLEAR;   
    }

    SDL_Event create_key_event(SDL_Scancode scancode, Uint16 mod) {
        SDL_Event event;

        event.type = SDL_KEYDOWN;
        event.key.timestamp = SDL_GetTicks();
        event.key.keysym.scancode = scancode;
        event.key.keysym.sym = SDL_GetKeyFromScancode(scancode);
        event.key.keysym.mod = mod;
        event.key.state = 1;

        return event;
    }

    SDL_Event create_textinput_event(SDL_Scancode scancode) {
        SDL_Event txt_in;
        txt_in.type = SDL_TEXTINPUT;

        try {
            const char* txt;

            if (app_workspace::get_instance()->lshift_flag) {
                try {
                    txt = custom_shift_keymap.at(scancode);
                    strcpy(txt_in.text.text, txt);
                    return txt_in;           
                } catch (std::out_of_range &ex) {
                    // no action, maybe log debug - if scancode isn't in shift_keymap, try looking in normal keymap
                }
            }
            
            txt = custom_keymap.at(scancode);

            // if the code is in cutom keymaps, use that instead of general SDL
            strcpy(txt_in.text.text, txt);
            return txt_in;           
        } catch (std::out_of_range &ex) {
            spdlog::debug("event_handler.cpp - Custom keymaps are missing request scancode ({0})", scancode);
        }
        
        strcpy(txt_in.text.text, SDL_GetScancodeName(scancode));
        // if scancode is a letter between A-Z and capslock is off, make lower case
        if (scancode >= SDL_SCANCODE_A && scancode <= SDL_SCANCODE_Z && !app_workspace::get_instance()->capslock_flag) {
            for (size_t i = 0; i < strlen(txt_in.text.text); i++) {
                txt_in.text.text[i] = std::tolower(txt_in.text.text[i]);
            }
        }

        return txt_in;
    }

    void push_event_to_sdl(SDL_Event* event, bool replace_cur_char) {
        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;

        spdlog::debug("event_handler.cpp - Pushing key event with scancode: {0} - {1}", event->key.keysym.scancode,
                     SDL_GetKeyName(SDL_GetKeyFromScancode(event->key.keysym.scancode)));
        
        // remove character so new one replaces it
        if (replace_cur_char && 
            kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT &&
            kbins->current_type == app_workspace_ns::kb_input_type::TEXT &&
            event->key.keysym.scancode != SDL_SCANCODE_RETURN)
        {
            SDL_Event backspace = create_key_event(SDL_SCANCODE_BACKSPACE, 0);
            SDL_PushEvent(&backspace);

            backspace.type = SDL_KEYUP;
            SDL_PushEvent(&backspace);
        }

        SDL_PushEvent(event); // pushes KEY_DOWN event

        if (kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT && 
            event->key.keysym.scancode != SDL_SCANCODE_RETURN &&
            event->key.keysym.scancode != SDL_SCANCODE_BACKSPACE) // backspace during input MUSTN'T send textinput event
        {
            SDL_Event txt_input = create_textinput_event(event->key.keysym.scancode);
            SDL_PushEvent(&txt_input); // pushes TEXTINPUT_EVENT of pressed key
        }

        event->type = SDL_KEYUP;
        event->key.state = 0;
        SDL_PushEvent(event); // pushes KEY_UP event
    }

    bool can_switch_modes() {
        screen_manager *scr_mgr = app_workspace::get_instance()->get_scr_mgr();
        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;

        if (kbins->current_mode != app_workspace_ns::kb_input_mode::NAVIGATION)
            return false;

        if (scr_mgr->activated)
            return false;

        // for now only gui_input can be activated
        if (!instanceof<gui_input>(scr_mgr->focused))
            return false;

        return true;
    }

    /**
     * @brief 
     * 
     * @param element 
     * @return int - temp: 1 when action is taken, 0 when function reaches end without any action
     */
    int focused_element_action(gui_element* element) {
        // for debug purposes
        if (instanceof<gui_button>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element BUTTON");
        } else if (instanceof<gui_combobox>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element COMBOBOX");
        } else if (instanceof<gui_direct>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element DIRECT");
        } else if (instanceof<gui_label>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element LABEL");
        } else if (instanceof<gui_observer>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element OBSERVER");
        } else if (instanceof<gui_selectable>(element)) {
            spdlog::debug("event_handler.cpp - action of focused element SELECTABLE");
        }else {
            spdlog::debug("event_handler.cpp - action of focused element UNKNOWN");
        }

        // handling button action - it needs to send TEXT_INPUT event as well
        if (instanceof<gui_button>(element) || instanceof<gui_combobox>(element) ||
            instanceof<gui_selectable>(element) || combo_expanded)
        {
            // on normal keyboard button callback is activated with "spacebar"
            // events are: KEYDOWN - SPACE, TEXTINPUT - SPACE, KEYUP - SPACE
            // only TEXTINPUT - SPACE might be enough to invoke callback (not tested!!!)

            spdlog::debug("event_handler.cpp - creating action event");

            SDL_Event ret = create_key_event(SDL_SCANCODE_SPACE, 0);
            SDL_PushEvent(&ret);

            SDL_Event txt_input = create_textinput_event(ret.key.keysym.scancode);
            SDL_PushEvent(&txt_input); // pushes TEXTINPUT_EVENT of pressed key

            ret.type = SDL_KEYUP;
            ret.key.state = 0;
            SDL_PushEvent(&ret); // pushes KEY_UP event

            if (instanceof<gui_combobox>(element) || combo_expanded) {
                combo_expanded = !combo_expanded; // if activating/ deactivating combo set this flag
                spdlog::debug("event_handler.cpp - setting combo_expanded flag to {0}", combo_expanded);
            }

            return 1;
        }

        return 0;
    }

    /**
     * @brief 
     * 
     * @return int - temp: 1 when action is performed, 0 when function reachec end without action
     */
    int handle_return_event() {
        screen_manager *scr_mgr = app_workspace::get_instance()->get_scr_mgr();
        app_workspace_ns::kb_input_state *kbins = app_workspace::get_instance()->kb_in_state.get();
        SDL_Event event;

        /*  how this works: if in navigation - enter on input enters input mode
                                             - enter on combobox/ button "activate event"
                            if in input mode - enter exits input mode
        */


        spdlog::debug("event_handler.cpp - handling RETURN raw event");

        if (kbins->current_mode == app_workspace_ns::kb_input_mode::INPUT) {
            spdlog::debug("event_handler.cpp - switching input mode from INPUT to NAVIGATION");

            event = create_key_event(SDL_SCANCODE_RETURN, 0);
            push_event_to_sdl(&event, false);
            kbins->current_mode = app_workspace_ns::kb_input_mode::NAVIGATION;

            // disable flags when exiting input mode (not necessary, does the same on input)
            app_workspace::get_instance()->lshift_flag = false;
            app_workspace::get_instance()->capslock_flag = false;

            return 1;
        }

        // see can_switch_modes for condition to enter INPUT mode
        if (can_switch_modes()) {
            spdlog::debug("event_handler.cpp - switching input mode from NAVIGATION to INPUT");

            // disable flags when entering input mode
            app_workspace::get_instance()->lshift_flag = false;
            app_workspace::get_instance()->capslock_flag = false;

            event = create_key_event(SDL_SCANCODE_RETURN, 0);
            push_event_to_sdl(&event, false);

            kbins->current_mode = app_workspace_ns::kb_input_mode::INPUT;
            // can_switch_modes() checks that focused is instanceof gui_input
            kbins->current_type = ((gui_input*) scr_mgr->focused)->get_in_type();
            spdlog::debug("event_handler.cpp - input type: {0}", kbins->current_type);

            return 1;
        }

        // when combo is expanded, an element IS focused, but it ISN'T any element from screen definition
        if (scr_mgr->focused || combo_expanded) {
            spdlog::debug("event_handler.cpp - calling focused element action");
            return focused_element_action(scr_mgr->focused);
        }
        
        spdlog::debug("event_handler.cpp - handling RETURN raw event with no action");
        return 0;
    }

    int handle_nav_event(kb_event* raw_event) {
        screen_manager *scr_mgr = app_workspace::get_instance()->get_scr_mgr();
        SDL_Event event;
        SDL_Scancode sc;

        // custom event handling here, doesn't need to be sent tro SDL

        sc = determine_scancode(raw_event->scancode);
        if (sc == SDL_SCANCODE_CLEAR)
            return 0;
        
        // 1-9 isn't send to SDL
        if (sc >= SDL_SCANCODE_1 && sc < SDL_SCANCODE_0) {
            uint8_t screen_no = strtol(SDL_GetScancodeName(sc), NULL, 10);
            scr_mgr->set_selected_screen(screen_no);
            return 1;
        }

        // 0  sends backspace - serves to go to previous screen
        if (sc == SDL_SCANCODE_BACKSPACE) {
            scr_mgr->pop_screen_from_previous();
            return 1; // not needed to send backspace during navigation to sdl (would be handled twice)
        }

        // create SDL event for keys other than 1-9: eg.: "arrow" UP, DOWN for nav
        event = create_key_event(sc);
        push_event_to_sdl(&event);

        return 1;
    }

    int handle_input_event(kb_event* raw_event) {
        std::unique_ptr<app_workspace_ns::kb_input_state> &kbins = app_workspace::get_instance()->kb_in_state;
        SDL_Event event;
        SDL_Scancode sc;

        // custom event handling here, doesn't need to be sent tro SDL

        // if input is text, determine if the key_should be cycled (e.g.: 2 = A/B/C)
        if (kbins->current_type == app_workspace_ns::kb_input_type::TEXT) {
            uint32_t delta_t = UINT32_MAX;
            // time since last event in us
            if (last_raw_event != nullptr) {
                delta_t = (raw_event->timestamp.tv_sec - last_raw_event->timestamp.tv_sec) * 1000000 +
                          (raw_event->timestamp.tv_usec - last_raw_event->timestamp.tv_usec);
                if (last_raw_event->scancode == raw_event->scancode && delta_t < 2000000) // 
                    key_cycle++;
                else
                    key_cycle = 0;
            } else {
                key_cycle = 0;
            }
         
            last_raw_event = raw_event;
        }

        sc = determine_scancode(raw_event->scancode, key_cycle);
        if (sc == SDL_SCANCODE_CLEAR)
            return 0;
        
        // those flags could be handled in key_cycle branch which would be a bit more efficient, but this is more readable
        // DO NOT create event for mod keys
        if (sc == SDL_SCANCODE_LSHIFT) {
            app_workspace::get_instance()->lshift_flag = !app_workspace::get_instance()->lshift_flag;
            return 1;
        }
        if (sc == SDL_SCANCODE_CAPSLOCK) {
            app_workspace::get_instance()->capslock_flag = !app_workspace::get_instance()->capslock_flag;
            return 1;
        }

        // doesn't do anything 
        Uint16 mod = 0;
        if (app_workspace::get_instance()->lshift_flag)    mod |= KMOD_LSHIFT;
        if (app_workspace::get_instance()->capslock_flag)  mod |= KMOD_CAPS;

        event = create_key_event(sc, mod);
        push_event_to_sdl(&event, key_cycle > 0);

        return 1;
    }

    int rfid_action_user_logged(rfid_event *event) {
        app_workspace *app_wrk = app_workspace::get_instance().get();
        user_workspace *usr_wrk = app_wrk->get_userspace();

        // if the rfid is the same as logged in user
        if (usr_wrk->compare_rfid_serials(&(event->tag)) == 0) { // 0 means equal
            spdlog::info("event_handler.cpp - RFID event handler: user with this RFID is already logged");
            // dont go to err screen (other than log, no action)
            return 1;                   
        }

        // currently logged user is pacient (0) and rfid doesnt match -> new user is trying to log in (regardless of new user role)
        if (usr_wrk->get_role() == 0) {
            spdlog::error("event_handler.cpp - RFID event handler: ");
            // TODO/ FIXME - maybe instead of error and manual log in, automatically log over
            app_wrk->set_err_screen_and_activate(get_localized_text("ERR_USER_LOGIN"),
                                                 get_localized_text("ERR_USER_ALREADY_LOGGED"));
            return 1;
        }

        if (usr_wrk->get_role() == 2 && app_wrk->get_scr_mgr()->get_selected_screen() == 6 &&
            app_wrk->s5_screen_6_indicator == 1) // if user is admin, active screen is 6 and content on screen is user addition
        { 
            app_wrk->s6_rfid_indicator = true;
            app_wrk->refresh_current_screen = true;
            rfid_reader::rfid_tag* tag_tmp = (rfid_reader::rfid_tag*) calloc(1, sizeof(rfid_reader::rfid_tag));
            memcpy(tag_tmp, &(event->tag), sizeof(rfid_reader::rfid_tag));
            app_wrk->s6_rfid_tag.reset(tag_tmp);
            return 0;
        }

        // this is reached only if role > 0 (employee/ admin). assuming the login is for subuser. subuser already exists, no place to log in new user
        if (usr_wrk->has_subuser()) {
            spdlog::error("event_handler.cpp - RFID event handler: already have user and subuser.");
            // can move to error screen with this, but rfid is usually loaded several times with one "press" on the reader
            // app_wrk->set_err_screen_and_activate(get_localized_text("ERR_USERS_FULL"),
            //                                      get_localized_text("ERR_USERS_FULL_DESC"));
            return 1;
        }

        app_wrk->log_in_user_rfid(&(event->tag), true);

        return 0;
    }

    int handle_rfid_event(rfid_event* event) {
        if (app_workspace::get_instance()->hx_measuring) {
            spdlog::debug("event_handler.cpp - Not handling keyboard events, because a hx measuring is ongoing.");
            return 1;
        }

        spdlog::info("event_handler.cpp - handling RFID event");
        app_workspace *app_wrk = app_workspace::get_instance().get();

        printf("RFID serial: ");
        for (int i = 0; i < event->tag.serial_size; i++) {
            printf("%02x(%d) ", event->tag.serial[i], event->tag.serial[i]);
        }
        printf("\n");

        if (!(event->flags & RFID_FLAGS_REPEATING)) {
            if (app_wrk->has_user()) {
                return rfid_action_user_logged(event);
            } else {
                if (app_wrk->get_scr_mgr()->get_selected_screen() == 1) {
                    app_wrk->log_in_user_rfid(&(event->tag), false);
                } else {
                    spdlog::error("event_handler.cpp - No user logged in. RFID login only possible from screen 1.");
                }
            }
        }

        return 0;
    }
}