#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <queue>
#include <chrono>
#include <spdlog/spdlog.h>
#include "rfid_reader.h"
#include "rc522.h"
#include "app_workspace.h"

#define REPEATING_FLAG_DELAY 500 //in ms
#define RFID_READ_INTERVAL 100000

namespace rfid_reader {
    std::queue<rfid_event> event_queue;
    std::mutex eq_mutex;

    std::unique_ptr<spi_config_t> spi_config;
    std::unique_ptr<SPI> my_spi = nullptr;
    bool reading_cards = true;


    void push_event(rfid_tag tag, uint8_t flags);

    int init(const char* device, uint8_t mode = 0, uint32_t speed = 1000000, uint16_t delay = 0, uint8_t bits_per_word = 8) {
        spi_config = std::unique_ptr<spi_config_t>(new spi_config_t());
        spi_config->mode = mode;
        spi_config->speed = speed;
        spi_config->delay = delay;
        spi_config->bits_per_word = bits_per_word;

        spdlog::info("rfid_reader.cpp - Initializing SPI RFID Reader (RC522)");

        my_spi = std::unique_ptr<SPI>(new SPI(device, spi_config.get()));

        if (!my_spi->begin()) {
            spdlog::critical("rfid_reader.cpp - Failed to start SPI communication for RC522");
            return 1;
        }

	    spdlog::info("rfid_reader.cpp - Successfully initialized SPI RFID Reader (RC522)");
        InitRc522(my_spi.get());
        return 0;
    }

    int init_from_conf() {
        app_workspace_ns::spi_config *config = &(app_workspace::get_instance()->main_config->rfid_conf);
        spdlog::info("rfid_reader.cpp - Initializing SPI RFID Reader (RC522) with config (enable debug to print config)");
        spdlog::debug("rfid_reader.cpp - dev: {0}"     , config->dev);
        spdlog::debug("rfid_reader.cpp - mode: {0}"    , config->config->mode);
        spdlog::debug("rfid_reader.cpp - bits_p_w: {0}", config->config->bits_per_word);
        spdlog::debug("rfid_reader.cpp - speed: {0}"   , config->config->speed);
        spdlog::debug("rfid_reader.cpp - delay: {0}"   , config->config->delay);

        my_spi = std::unique_ptr<SPI>(new SPI(config->dev.c_str(), config->config.get()));

        if (!my_spi->begin()) {
            spdlog::critical("rfid_reader.cpp - Failed to start SPI communication for RC522");
            return 1;
        }

	    spdlog::info("rfid_reader.cpp - Successfully initialized SPI RFID Reader (RC522)");
        InitRc522(my_spi.get());
        return 0;
    }

    void print_card_serial(rfid_tag *tag) {
        spdlog::debug("rfid_reader.cpp - Read RFID tag. Serial:");
        // printf isn't very elegant, but this is a debug function anyway
        for (int i = 0; i < tag->serial_size; i++) {
            printf("%d", tag->serial[i]);
        }
        printf("\n");
    }

    void start_reading_cards() {
        uint8_t flags;
        int rv;
        rfid_tag tag;
        std::chrono::time_point<std::chrono::high_resolution_clock> last_read =
            std::chrono::high_resolution_clock::now();
        std::chrono::time_point<std::chrono::high_resolution_clock> current_read;

        while (reading_cards) {
            flags = RFID_FLAGS_NONE;
            rv = read_tag(&tag);
            current_read = std::chrono::high_resolution_clock::now();

            if (rv != TAG_OK) {
                spdlog::error("rfid_reader.cpp - read_tag() finished but returned error state");
            } else {
                spdlog::info("rfid_reader.cpp - Successfully read RFID tag. Pushing into RFID event queue.");

                std::chrono::duration<double, std::milli> diff = last_read - current_read;
                if (diff.count() > REPEATING_FLAG_DELAY) {
                    flags |= RFID_FLAGS_REPEATING;
                }
                //print_card_serial(&tag);
                push_event(tag, flags);
            }

            last_read = current_read; //maybe change its to ::now()
        }
    }

    void clean() {
        spdlog::info("rfid_reader.cpp - Cleaning SPI RFID Reader (RC522)");
        reading_cards = false;
        // my_spi is unique_ptr, should destruct automatically
    }

    char detect_tag(uint16_t *tag_type, uint8_t *buff) {
        char status;
        status = PcdRequest(PICC_REQALL, buff);
    
        if (status  == TAG_OK) {
            *tag_type = buff[0] << 8 | buff[1];
        }
    
        return status;
    }
    
    char read_tag_serialn(uint8_t *serial, uint8_t *serial_size, uint8_t *ser_ack, uint8_t *buff) {
        /* get basic UID from card */
        if (PcdAnticoll(PICC_ANTICOLL1, buff) != TAG_OK) return TAG_ERR;
    
        /* select the card */
        if (PcdSelect(PICC_ANTICOLL1, buff, ser_ack) != TAG_OK) return TAG_ERR;
        
        if (buff[0]== 0x88) {       // 0x88 is CT = Cascade Tag (means next level to get UID)
    
            //copy UID0, UID1, UID2
            memcpy(serial, &buff[1],3);
            
            // 0x95 => get cascade level 2 and select
            if (PcdAnticoll(PICC_ANTICOLL2, buff)!= TAG_OK) return TAG_ERR;
            
            if (PcdSelect(PICC_ANTICOLL2, buff, ser_ack) != TAG_OK) return TAG_ERR;
    
            if (buff[0]==0x88) 
            {
                // add UID3, UID4 and UID5
                memcpy(serial + 3, &buff[1], 3);
               
                // 0x97 => get even more extended serial number
                
                if (PcdAnticoll(PICC_ANTICOLL3, buff) != TAG_OK)  return TAG_ERR;
    
                if (PcdSelect(PICC_ANTICOLL3, buff, ser_ack) != TAG_OK) return TAG_ERR;
    
                // add UID6, UID7, UID8 UID9
                memcpy(serial + 6, buff, 4);
                *serial_size = 10;        // 10 byte UID (triple size UID) PSS
            } else {
                // add UID3, UID4, UID5 UID6
                memcpy(serial + 3, buff, 4);
                /* 7 byte UID (double size) Mifare Ultralight, Ultralight C, Desfire
                 * Desfire EV1 and Mifare Plus (7 B UID)
                 */
                *serial_size = 7;         // 7 byte Unique ID (double size UID)
            }
        } else {
            //copy UID0, UID1, UID2, UID3
            memcpy(serial, &buff[0], 4);
            /* 4 byte non-unique ID (single byte UID) - discontinued end 2010
             * were set at manufacturing.
             * used on Mifare Classic 1K, Mifare Classic 4K and some Mifare Plus versions
             */
            *serial_size = 4;
        }
    
        return TAG_OK;
    }
    
    int read_tag(rfid_tag *tag) {
        int status;
        uint8_t buff[MAXRLEN];
        //rfid_tag tag;
        //p_printf(BLUE,"Hold your card for the reader "); // debug "reding card"
    
        do {
            status = detect_tag(&(tag->tag_type), buff);
    
            if (status == TAG_NOTAG) { // no card - wait
                usleep(RFID_READ_INTERVAL);
                continue;
            } else if (status != TAG_OK && status != TAG_COLLISION) {
                // issue with tag reading ?
                spdlog::debug("rfid_reader.cpp - Issue reading tag, trying again...");
                usleep(RFID_READ_INTERVAL / 2);
                continue;
            }
    
            //maybe set buff mem to 0?
    
            // read serial number & SAK
            status = read_tag_serialn(tag->serial, &(tag->serial_size), tag->ser_ack, buff);
    
        } while (status != TAG_OK && reading_cards);
    
        spdlog::debug("rfid_reader.cpp - Found card");
    
        /* determine memory details
         * but this is very cryptic and hard to exactly discover
         * SAK = Select acknowledgement has card details in it but still...
         * nearly mission impossible
         */
    
        tag->max_blocks = 0;
    
        switch (tag->tag_type) {
            case 0x4400:          
    
                if (tag->ser_ack[0] == 0x0) {
                    spdlog::debug("rfid_reader.cpp - Ultralight card");
                    spdlog::debug("rfid_reader.cpp - NOT supported by this program");
                } else if (tag->ser_ack[0] == 0x8) {
                    spdlog::debug("rfid_reader.cpp - Mifare Classic(1K) or PLUS card (2K)");
                    spdlog::debug("rfid_reader.cpp - In case of PLUS card, we will only read the first 1K (64 blocks)");
                    tag->max_blocks=64;	
                    tag->page_step=1;       // increment to read the card
                } else if (tag->ser_ack[0] == 0x9) {// 20 blocks (5 sectors x 4 blocks) x 16 bytes
                    spdlog::debug("rfid_reader.cpp - Mifare MINI(0.3K)");
                    tag->max_blocks=20;
                    tag->page_step=1;       // increment to read the card
                } else if (tag->ser_ack[0] == 0x18) {
                    spdlog::debug("rfid_reader.cpp - PLUS card (4K)");
                    spdlog::debug("rfid_reader.cpp - Can only access the first 128 blocks (NOT last 8 sectore with 16 blocks)");
                    tag->max_blocks=128;
                    tag->page_step=1;       // increment to read the card
                }
                break;
    
            case 0x0400:		
                if (tag->ser_ack[0] == 0x8) { // 64 blocks of 16 bytes / Ul card
                    spdlog::debug("rfid_reader.cpp - Mifare Classic(1K) or SmartMX with MIFARE 1K emulation");
                    tag->max_blocks=64;
                    tag->page_step=1;       // increment to read the card
                } else if (tag->ser_ack[0] == 0x9) { // 20 blocks (5 sectors x 4 blocks) x 16 bytes
                    spdlog::debug("rfid_reader.cpp - Mifare MINI(0.3K)");
                    tag->max_blocks=20;
                    tag->page_step=1;       // increment to read the card
                }
                break;
    
            case 0x0200:		
                if (tag->ser_ack[0] == 0x18) { // 64 blocks of 16 bytes / Ul card
                    spdlog::debug("rfid_reader.cpp - Mifare Classic(4K) or SmartMX with MIFARE 4K emulation");
                    spdlog::debug("rfid_reader.cpp - Can only access the first 128 blocks (NOT last 8 sectore with 16 blocks)");
                    tag->max_blocks=127;
                    tag->page_step=1;       // increment to read the card
                } else if (tag->ser_ack[0] == 0x9) { // 20 blocks (5 sectors x 4 blocks) x 16 bytes
                    spdlog::debug("rfid_reader.cpp - Mifare MINI(0.3K)");
                    tag->max_blocks=20;
                    tag->page_step=1;       // increment to read the card
                }
                break;
        }
    
        if (tag->max_blocks == 0) {
            spdlog::error("rfid_reader.cpp - Read card, BUT card type not known");
            //PcdHalt();
            // set card reading loop var to false - stop reading cards?
        }
    
        return status;
    }

    /** This function creates a RFID event and pushes it into the event queue */
    void push_event(rfid_tag tag, uint8_t flags) {
        rfid_event event = {};

        event.tag = tag;
        event.flags = flags;
        event_queue.push(event);
    }

    rfid_event* poll_event() {
        if (event_queue.size() <= 0)
            return NULL;

        // probably don't need to lock thread for poll, but better safe than sorry
        std::lock_guard<std::mutex> eq_guard(eq_mutex);

        rfid_event* res = &event_queue.front();
        event_queue.pop();

        return res;
    }

    void clear_event_queue() {
        for (size_t i = 0; i < event_queue.size(); i++) {
            event_queue.pop();
        }
    }
}
