#include "app_workspace.h"

#include <stdio.h>
#include <spdlog/spdlog.h>
#include "imgui.h"
#include "localisation.h"
#include "gui_label.h"

#include "unistd.h"


/**
 * @brief This function is started in a new thread for reading HX711 values
 * 
 */
void threaded_values_read();
/**
 * @brief This function is started in a new thread for taring HX711
 * 
 */
void threaded_tare();
/**
 * @brief This function is started in a new thread for calibrating HX711
 * 
 */
void threaded_calibration();
/**
 * @brief This function is started in a new thread for reading HX711 values in a loop. This
 * reading is done in intervals.
 * 
 */
void threaded_values_read_looped();
/**
 * @brief When an HX711 reading is started, when creating new thread, this thread is created as well.
 *  This thread serves for measuring values written by HX711.
 * 
 * @param msring When this is set to false, watchdog is stopped
 * @param watcher Values, that are watched
 */
void threaded_measuring_watchdog(bool *msring, std::vector<HX711::Value> **watcher);

// init instance
std::unique_ptr<app_workspace> app_workspace::instance;

std::unique_ptr<app_workspace> &app_workspace::get_instance() {
    if (!instance.get()) {
        spdlog::debug("app_workspace.cpp - Instantiating app_workspace");
        instance = std::unique_ptr<app_workspace>(new app_workspace());
    }
    return instance;
}

/**
 * One of the most importat classes in the whole project. Holds variables that define the state of the application 
 * and also provides logic for some of the variables held there. For example the hx711 controller is held and
 * used from this class.
 */
app_workspace::app_workspace() {
}

void app_workspace::init_labels() {
    this->s6_role_labels[0] = get_localized_text("GT_ROLE_1");
    this->s6_role_labels[1] = get_localized_text("GT_ROLE_2");
    this->s6_role_labels[2] = get_localized_text("GT_ROLE_3");

    this->s6_month_labels[0] = get_localized_text("GT_MONTH_01");
    this->s6_month_labels[1] = get_localized_text("GT_MONTH_02");
    this->s6_month_labels[2] = get_localized_text("GT_MONTH_03");
    this->s6_month_labels[3] = get_localized_text("GT_MONTH_04");
    this->s6_month_labels[4] = get_localized_text("GT_MONTH_05");
    this->s6_month_labels[5] = get_localized_text("GT_MONTH_06");
    this->s6_month_labels[6] = get_localized_text("GT_MONTH_07");
    this->s6_month_labels[7] = get_localized_text("GT_MONTH_08");
    this->s6_month_labels[8] = get_localized_text("GT_MONTH_09");
    this->s6_month_labels[9] = get_localized_text("GT_MONTH_10");
    this->s6_month_labels[10] = get_localized_text("GT_MONTH_11");
    this->s6_month_labels[11] = get_localized_text("GT_MONTH_12");

    this->s6_unit_sel_labels[0] = get_localized_text("GT_G_UNIT");
    this->s6_unit_sel_labels[1] = get_localized_text("GT_KG_UNIT");
    
    this->unit_sel_labels[0] = get_localized_text("GT_MG_UNIT");
    this->unit_sel_labels[1] = get_localized_text("GT_G_UNIT");
    this->unit_sel_labels[2] = get_localized_text("GT_KG_UNIT");

    this->hx_measure_opts[0] = get_localized_text("SCREEN_2_CB1_SAMPLES");
    this->hx_measure_opts[1] = get_localized_text("SCREEN_2_CB1_TIMEOUT");

    this->s3_filter_opt_labels[0] = get_localized_text("SCREEN_3_CB1_LARGER");
    this->s3_filter_opt_labels[1] = get_localized_text("SCREEN_3_CB1_SMALLER");
}

int app_workspace::init_hx711_controller() {
    try {
        HX711::Rate rate = HX711::Rate::HZ_80;
        if (this->main_config->hx_conf.rate == 80) {
            rate = HX711::Rate::HZ_80;
        } else if (this->main_config->hx_conf.rate == 10) {
            rate = HX711::Rate::HZ_10;
        } else {
            rate = HX711::Rate::OTHER;
        }

        // hx711_controller = std::make_unique<HX711::AdvancedHX711>(
        hx711_controller = std::make_unique<custom_hx711>(
            this->main_config->hx_conf.data_pin,
            this->main_config->hx_conf.clock_pin,
            this->main_config->hx_conf.ref_unit,
            this->main_config->hx_conf.offset,
            rate
        );

        return 0;
    } catch (HX711::GpioException &ex) { 
        std::string what = ex.what();
        std::string err = "app_workspace.cpp - hx711 initialization failed. Reason: " + what;
        spdlog::critical(err);

        return 1;
    }
}

void app_workspace::hx711_tare() {
    this->hx_measuring = true;
    this->refresh_current_screen = true;

    this->s6_calibration_finished = false; // running tare means there is not calibration

    // using calib_thread because tare can't be run at the same time as calibration
    this->calib_thread.reset(new std::thread(threaded_tare));
    this->calib_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway
}

void app_workspace::hx711_calibrate() {
    this->hx_measuring = true;
    this->refresh_current_screen = true;

    // this->s6_tare_complete = false; // when calibration is run, tare is reset (cant know if weight is stil zeroed)

    this->calib_thread.reset(new std::thread(threaded_calibration));
    this->calib_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway
}

int app_workspace::hx711_measure() {
    if (this->hx711_controller.get() == nullptr) {
        set_err_screen_and_activate(get_localized_text("ERR_HX_ERROR"),
                                    get_localized_text("ERR_HX_NOT_INIT_DESC"));
        return 5;
    }

    if (!this->user_logged || this->userspace.get() == nullptr || !this->userspace->has_user()) {
        set_err_screen_and_activate(get_localized_text("ERR_NO_USER_LOGIN"),
                                    get_localized_text("ERR_NO_USER_LOGIN_DESC"));
        return 1; // can't measure if no user is logged in
    }

    if (this->selected_hx_measure_opt == 0) {
        if (this->samples_in < S2_MIN_SAMPLES_IN || this->samples_in > S2_MAX_SAMPLES_IN) {
            set_err_screen_and_activate(get_localized_text("ERR_S2_INPUT"),
                                        get_localized_text("ERR_S2_SAMPLES_IN_ERR"));
            spdlog::error("app_workspace.cpp - samples_in: {0}", this->samples_in);
            this->clear_s2_buffers();
            return 2;
        }
    } else if (this->selected_hx_measure_opt == 1) {
        if (this->timeout_in < S2_MIN_TIMEOUT_IN || this->timeout_in > S2_MAX_TIMEOUT_IN) {
            set_err_screen_and_activate(get_localized_text("ERR_S2_INPUT"),
                                        get_localized_text("ERR_S2_TIMEOUT_IN_ERR"));
            spdlog::error("app_workspace.cpp - timeout_in: {0}", this->timeout_in);
            this->clear_s2_buffers();
            return 2;
        }
    } else {
        set_err_screen_and_activate(get_localized_text("ERR_S2_UNKNOWN"),
                                    get_localized_text("ERR_S2_UNKNOWN_DESC"));
        return 3;
    }

    // reset this flag, if the measurement is stopped in the moment its finishing, 
    // this will stay set to true and cause the new measurement not to save
    this->hx_not_finished_yet = false;
    this->hx_continuous = false;

    this->hx_measuring = true;
    this->refresh_current_screen = true;

    this->measuring_thread.reset(new std::thread(threaded_values_read));
    this->measuring_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway

    return 0;
}

int app_workspace::hx711_continuous_measure() {
    if (this->hx711_controller.get() == nullptr) {
        set_err_screen_and_activate(get_localized_text("ERR_HX_ERROR"),
                                    get_localized_text("ERR_HX_NOT_INIT_DESC"));
        return 1;
    }

    if (!this->user_logged || this->userspace.get() == nullptr || !this->userspace->has_user()) {
        set_err_screen_and_activate(get_localized_text("ERR_NO_USER_LOGIN"),
                                    get_localized_text("ERR_NO_USER_LOGIN_DESC"));
        return 2; // can't measure if no user is logged in
    }

    // reset this flag, if the measurement is stopped in the moment its finishing, 
    // this will stay set to true and cause the new measurement not to save
    this->hx_not_finished_yet = false;
    this->hx_measuring = true;
    this->hx_continuous = true;
    this->refresh_current_screen = true;

    this->measuring_thread.reset(new std::thread(threaded_values_read_looped));
    this->measuring_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway

    return 0;
}

std::vector<HX711::Value> app_workspace::hx711_values_samples(int samples) {
    if (!this->hx_continuous)
        spdlog::info("app_workspace.cpp - Starting HX711 measuring for {0} samples.", samples);
    std::mutex mtx;

    bool measuring = true;
    std::vector<HX711::Value> *watcher = nullptr;

    this->measuring_watchdog.reset(new std::thread(threaded_measuring_watchdog, &measuring, &watcher));
    // this->measuring_watchdog->detach(); // join after measuring would be preferrable, but that is potential deadlock

    spdlog::debug("STARTING measuring");

    mtx.lock();
    m_start = std::chrono::high_resolution_clock::now();
    mtx.unlock();
    //FIXME this might throw exception if it fails
    std::vector<HX711::Value> res = hx711_controller->getValues(samples, &watcher);
    mtx.lock();
    m_end = std::chrono::high_resolution_clock::now();
    mtx.unlock();

    mtx.lock();
    measuring = false; // needed to stop watchdog thread!!
    mtx.unlock();

    this->measuring_watchdog->join();

    spdlog::debug("ENDING measuring");

    return res;
}

std::vector<HX711::Value> app_workspace::hx711_values_timeout(int seconds) {
    spdlog::info("app_workspace.cpp - Starting HX711 measuring for {0} seconds.", seconds);
    std::mutex mtx;

    bool measuring = true;
    std::vector<HX711::Value> *watcher = nullptr;

    this->measuring_watchdog.reset(new std::thread(threaded_measuring_watchdog, &measuring, &watcher));
    // this->measuring_watchdog->detach(); // join after measuring would be preferrable, but that is potential deadlock

    spdlog::debug("STARTING measuring");

    mtx.lock();
    m_start = std::chrono::high_resolution_clock::now();
    mtx.unlock();
    //FIXME this might throw exception if it fails (same as above)
    std::vector<HX711::Value> res = hx711_controller->getValues(std::chrono::seconds(seconds), &watcher);
    mtx.lock();
    m_end = std::chrono::high_resolution_clock::now();
    mtx.unlock();

    mtx.lock();
    measuring = false; // needed to stop watchdog thread!!
    mtx.unlock();

    this->measuring_watchdog->join();

    spdlog::debug("ENDING measuring");

    // spdlog::info("app_workspace.cpp - HX711 finished taking raw values.");

    return res;
}

screen_manager* app_workspace::get_scr_mgr() {
    return this->scr_mgr;
    //return nullptr;
}

/**
 * @brief This function should only be used once in the program.
 * Function is required, because screen_manager cannot be initialized
 * during construction due to cycling.
 * 
 * @param scr_mgr 
 */
void app_workspace::set_scr_mgr(screen_manager *scr_mgr) {
    this->scr_mgr = scr_mgr;
}

void app_workspace::set_err_screen_and_activate(const char* title, const char* label) {
    extern gui_label *err_title;
    extern gui_label *err_desc;
    err_title->set_label(title);
    err_desc->set_label(label);
    this->scr_mgr->set_selected_screen(0);
}

void app_workspace::use_font_size(app_workspace_ns::font_size font_size) {
    ImGui::PushFont(this->loaded_fonts.at(font_size));
    this->font_push_cnt++;
    cur_font_size = font_size;
}

app_workspace_ns::font_size app_workspace::get_font_size() {
    return this->cur_font_size;
}

int app_workspace::load_font_size(app_workspace_ns::font_size key, float size_px) {
    spdlog::info("app_workspace.cpp - Loading font with size {0}", size_px);

    ImGuiIO &io = ImGui::GetIO();
    std::string font_path = this->main_config->fonts_path + this->main_config->font;
    ImFont* font = io.Fonts->AddFontFromFileTTF(font_path.c_str(), size_px);
    
    if (font == NULL) {
        // FIXME - maybe critical
        spdlog::error("app_workspace.cpp - Failed to load font size {0}!" , size_px);
        return 1;
    }

    this->loaded_fonts.emplace(key, font);

    return 0;
}

int app_workspace::log_in_user_rfid(rfid_reader::rfid_tag *tag, bool loading_sub) {
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));
    int rv;

    spdlog::info("app_workspace.cpp - Loging in user through RFID. Querrying DB...");
    
    rv = db_conn->query_user_data(usr, nullptr, tag->serial, tag->serial_size);
    if (rv) {
        spdlog::error("app_workspace.cpp - Failed to login user. db_driver log should provide more information.");
        free(usr);
        return 1;
    }

    rv = db_conn->query_measurement_headers(usr);
    if (rv) {
        spdlog::error("app_workspace.cpp - Failed to retrieve user measurement headers.");
        // free(usr);
        // return 1;
    }

    if (loading_sub) {
        if (userspace.get() == nullptr || !userspace->has_user()) {
            spdlog::error("app_workspace.cpp - Cannot login \"subuser\", because user workspace isn't initialized."
                          " User isn't logged in? flag: user_logged: {0}", user_logged);
            free(usr);
            return 1;
        }

        if (userspace->get_user_id() == usr->id) {
            this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                              get_localized_text("ERR_SUBUSER_SAME_USER"));
            free(usr);
            return 1;
        }

        this->userspace->subuser.reset(usr);
        this->scr_mgr->refresh_all_screns();
        spdlog::info("app_workspace.cpp - Finished logging in subuser through RFID.");

        return 0;
    }

    // loading_sub is false
    userspace.reset(new user_workspace(usr));
    user_logged = true;
    scr_mgr->refresh_all_screns();

    spdlog::info("app_workspace.cpp - Finished logging in user through RFID.");

    return 0;
}

int app_workspace::verify_user_cred() {
    int rv;
    user_cred credentials;

    rv = db_conn->query_user_credentials(&credentials, usr_name);
    if (rv == 2) {
        this->set_err_screen_and_activate(get_localized_text("ERR_USER_CRED"),
                                          get_localized_text("ERR_USER_NOT_FOUND_DESC"));
        return 1;
    }

    if (rv) {
        this->set_err_screen_and_activate(get_localized_text("ERR_USER_CRED"),
                                          get_localized_text("ERR_USER_CRED_GENERAL_DESC"));
        return 1;
    }

    if (strcmp(credentials.passwd, usr_passwd)) {
        // FIXME - "credentials are wrong" - maybe do "wrong password" (but thats weaker security)
        this->set_err_screen_and_activate(get_localized_text("ERR_USER_CRED"),
                                          get_localized_text("ERR_USER_CRED_WRONG_DESC"));
        return 1;
    }

    return 0;
}

//int app_workspace::log_in_user_cred() {
void app_workspace::log_in_user_cred() {
    if (strlen(usr_name) <= 0 || strlen(usr_passwd) <= 0) {
        this->set_err_screen_and_activate(get_localized_text("ERR_USER_LOGIN"),
                                          get_localized_text("ERR_USER_CREDS_NOT_ENTERED"));
        return;
    }

    if (verify_user_cred()) {
        spdlog::error("app_workspace.cpp - User login. Failed to verify credentials");
        return;
    }

    int rv;
    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));
    
    rv = db_conn->query_user_data(usr, usr_name);
    if (rv) {
        // this hould be unreachable, because user was already checked for credentials -> user exists, sql can fail
        spdlog::error("app_workspace.cpp - Failed to login user. db_driver log should provide more information.");
        //return 1;
        return;
    }

    rv = db_conn->query_measurement_headers(usr);
    if (rv) {
        spdlog::error("app_workspace.cpp - Failed to retrieve user measurement headers.");
        // return;
    }

    // loading_sub is false
    userspace.reset(new user_workspace(usr));
    user_logged = true;
    clear_s1_buffers();
    scr_mgr->refresh_all_screns();

    //return 0;
}

void app_workspace::login_subuser() {
    // not checking has_user(), because this action is currently only available from user logged in screen

    if (strlen(subusr_name) <= 0) {
        this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                          get_localized_text("ERR_SUBUSER_INPUT_EMPTY"));
        return;
    }

    int rv = db_conn->can_login_subuser(subusr_name);
    if (rv == 2) {
        this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                          get_localized_text("ERR_SUBUSER_NOT_FOUND"));
        return;
    } else if (rv == 3) {
        this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                          get_localized_text("ERR_SUBUSER_ROLE_TOO_HIGH"));
        return;
    }
    if (rv) {
        this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                          get_localized_text("ERR_SUBUSER_FAILED_TO_FETCH"));
        return;
    }

    user_cont *usr = (user_cont*) calloc(1, sizeof(user_cont));
    
    rv = db_conn->query_user_data(usr, subusr_name);
    if (rv) {
        spdlog::error("app_workspace.cpp - Failed to login user. db_driver log should provide more information.");
        return;
    }

    rv = db_conn->query_measurement_headers(usr);
    if (rv) {
        spdlog::error("app_workspace.cpp - Failed to retrieve user measurement headers.");
        return;
    }

    if (userspace->get_user_id() == usr->id) {
        this->set_err_screen_and_activate(get_localized_text("ERR_SUBUSER_LOGIN"),
                                          get_localized_text("ERR_SUBUSER_SAME_USER"));
        free(usr);
        return;
    }

    this->userspace->subuser.reset(usr);
    this->scr_mgr->refresh_all_screns();
}

void app_workspace::logout_subuser() {
    clear_s1_buffers();
    this->userspace->subuser.release();
    this->userspace->measured.release();
    this->userspace->picked.release();
    this->scr_mgr->refresh_all_screns();
}

void app_workspace::logoff_user() {
    this->logout_subuser();
    // clear_s1_buffers(); // these should be cleared after login, but just to be sure
    userspace.release();
    user_logged = false;
    scr_mgr->refresh_all_screns();
    scr_mgr->clear_previous_screen_stack();
}

int app_workspace::register_new_user() {
    if (strlen(this->s6_uname_in)    <= 0 ||
        strlen(this->s6_passwd_in)   <= 0 ||
        strlen(this->s6_name_in)     <= 0 ||
        strlen(this->s6_lastname_in) <= 0 ||
        this->s6_year_in <= 0 ||
        this->s6_day_in  <= 0)
    {
        this->set_err_screen_and_activate(get_localized_text("ERR_S6_USR_ADD_INPUT_ERR"),
                                          get_localized_text("ERR_S6_USR_ADD_INPUT_DESC"));
        return 1;
    }

    user_cont *usr_tmp = (user_cont*) calloc(1, sizeof(user_cont));
    std::unique_ptr<user_cont> usr(usr_tmp);

    int rv = this->db_conn->is_username_available(this->s6_uname_in);
    if (rv == 3) {
        this->set_err_screen_and_activate(get_localized_text("ERR_S6_USR_ADD"),
                                          get_localized_text("ERR_S6_USR_NAME_TAKEN"));
        return 1;
    } else if (rv) {
        spdlog::error("screen_definitions.cpp - Failed to verify username availability");
        this->set_err_screen_and_activate(get_localized_text("ERR_DB_ERROR"),
                                          get_localized_text("ERR_DB_UNAME_CHECK_FAIL"));
        return 1;
    } else {
        spdlog::debug("screen_definitions.cpp - Username is available.");
    }

    if (this->s6_rfid_indicator) {
        rv = this->db_conn->is_rfid_serial_available(this->s6_rfid_tag->serial, this->s6_rfid_tag->serial_size);

        if (rv == 3) {
            this->set_err_screen_and_activate(get_localized_text("ERR_S6_USR_ADD"),
                                              get_localized_text("ERR_S6_USR_RFID_TAKEN"));
            return 1;
        } else if (rv) {
            spdlog::error("screen_definitions.cpp - Failed to verify username availability");
            this->set_err_screen_and_activate(get_localized_text("ERR_DB_ERROR"),
                                              get_localized_text("ERR_DB_RFID_CHECK_FAIL"));
            return 1;
        } else {
            spdlog::debug("screen_definitions.cpp - RFID is available.");
            memcpy(usr->rfid_serial_bin, this->s6_rfid_tag->serial, this->s6_rfid_tag->serial_size);
            usr->rfid_ser_len = this->s6_rfid_tag->serial_size;
        }
    }

    std::time_t ct_tmp = std::time(nullptr);
    std::tm *current_time = std::localtime(&ct_tmp);

    // tm_year is "current_year - 1900"
    if (this->s6_year_in > (current_time->tm_year + 1900) || this->s6_year_in < YEAR_MIN) {
        // spdlog::error("screen_definitions.cpp - Year out of range {0} (current_year: {1})", 
        //     this->s6_year_in, (current_time->tm_year + 1900));
        this->set_err_screen_and_activate(get_localized_text("ERR_DOB"),
                                          get_localized_text("ERR_DOB_YEAR_OOR"));
        return 1;
    }
    if (this->s6_day_in > DAY_MAX || this->s6_day_in < DAY_MIN) {
        spdlog::error("screen_definitions.cpp - Year out of range");
        this->set_err_screen_and_activate(get_localized_text("ERR_DOB"),
                                          get_localized_text("ERR_DOB_DAY_OOR"));
        return 1;
    }

    char buf[64];
    sprintf(buf, "%04d-%02d-%02d", this->s6_year_in, this->s6_selected_month + 1, this->s6_day_in);

    std::tm dob;
    memset(&dob, 0, sizeof(std::tm));
    strptime(buf, "%Y-%m-%d", &dob);

    usr->role     = this->s6_selected_role;
    usr->username = this->s6_uname_in;
    usr->name     = this->s6_name_in;
    usr->lastname = this->s6_lastname_in;
    if (strlen(this->s6_desc_in) > 0) {
        usr->description = this->s6_desc_in;
    }
    usr->date_of_birth = dob;

    rv = this->db_conn->insert_user(usr.get(), this->s6_passwd_in);
    if (rv) {
        spdlog::error("screen_definitions.cpp - Failed to execute user insert.");
        this->set_err_screen_and_activate(get_localized_text("ERR_DB_ERROR"),
                                          get_localized_text("ERR_DB_INSERT_USER_FAIL"));
        return 1;
    } else {
        this->clear_s6_buffers();
        spdlog::info("screen_definitions.cpp - Successfully added new user.");
        this->scr_mgr->set_selected_screen(5);
    }

    return 0;
}

bool app_workspace::is_debug_screens() {
    return main_config->debug_screens;
}


bool app_workspace::has_user() {
    return this->user_logged && this->userspace.get() != nullptr && this->userspace->has_user();
}

user_workspace* app_workspace::get_userspace() {
    if (has_user())
        return this->userspace.get();
    else
        return nullptr;
}

int app_workspace::open_db_connection() {
    db_conn.reset(new db_driver(&(main_config->db_conf)));
    return db_conn->is_open() ? 0 : 1;
}

void app_workspace::clear_s1_buffers() {
    memset(usr_name, 0, DEF_BUFF_SIZE);
    memset(subusr_name, 0, DEF_BUFF_SIZE);
    memset(usr_passwd, 0, DEF_BUFF_SIZE_SMALL);
}

void app_workspace::clear_s2_buffers() {
    this->samples_in = S2_MIN_SAMPLES_IN;
    this->timeout_in = S2_MIN_TIMEOUT_IN;
}

void app_workspace::clear_s6_buffers() {
    this->s6_ref_mass_in = 0;
    this->s6_samples_in = S6_MIN_SAMPLES_IN;
    this->s6_selected_unit_option = 0;
    this->s6_day_in = DAY_MIN;
    this->s6_year_in = YEAR_MIN;
    this->s6_calibration_finished = false;
    this->s6_tare_complete = false;
    this->s6_rfid_indicator = false;
    this->s6_rfid_tag.release();
    this->s6_selected_role = 0;
    this->s6_selected_month = 0;
    memset(this->s6_lastname_in, 0, DEF_BUFF_SIZE_SMALL);
    memset(this->s6_uname_in, 0, DEF_BUFF_SIZE_SMALL);
    memset(this->s6_name_in, 0, DEF_BUFF_SIZE);
    memset(this->s6_passwd_in, 0, DEF_BUFF_SIZE);
    memset(this->s6_desc_in, 0, DEF_BUFF_SIZE_EXTRA);
    
}

std::vector<double> app_workspace::normalize_raw_values(std::vector<HX711::Value>& raw_vals) {
    std::vector<double> res;

    for (size_t i = 0; i < raw_vals.size(); i++) {
        res.push_back(this->hx711_controller->normalise(raw_vals[i]));
    }

    return res;
}

/*********************************************************************************************************************/
/*       Debug/ test fuctinos                                                                                        */
/*********************************************************************************************************************/


void app_workspace::print_config() {
    spdlog::debug("========== (debug) config file start ==========");
    spdlog::debug("general settings:");
    spdlog::debug("\tdebug_screens: {0}",   main_config->debug_screens);
    spdlog::debug("\tfonts_path: {0}",      main_config->fonts_path);
    spdlog::debug("\tfont: {0}",            main_config->font);
    spdlog::debug("\tlang_path: {0}",       main_config->lang_path);
    spdlog::debug("\tavailable_langs:");
    for (size_t i = 0; i < main_config->lang_options.size(); i++) {
        spdlog::debug("\t\topt {0}: {1}", i, main_config->lang_options[i]);
    }
    //spdlog::debug("\tavailable_langs: {0}", main_config->lang_options);
    spdlog::debug("\tdefault_lang: {0}",    main_config->lang_default);
    spdlog::debug("\tcurrent_lang: {0}",    main_config->lang_current);
    spdlog::debug("spi keyboard settings:");
    spdlog::debug("\tdev: {0}",             main_config->keyb_conf.dev);
    spdlog::debug("\tmode: {0}",            main_config->keyb_conf.config->mode);
    spdlog::debug("\tbits: {0}",            main_config->keyb_conf.config->bits_per_word);
    spdlog::debug("\tspeed: {0}",           main_config->keyb_conf.config->speed);
    spdlog::debug("\tdelay: {0}",           main_config->keyb_conf.config->delay);
    spdlog::debug("spi rfid settings:");
    spdlog::debug("\tdev: {0}",             main_config->rfid_conf.dev);
    spdlog::debug("\tmode: {0}",            main_config->rfid_conf.config->mode);
    spdlog::debug("\tbits: {0}",            main_config->rfid_conf.config->bits_per_word);
    spdlog::debug("\tspeed: {0}",           main_config->rfid_conf.config->speed);
    spdlog::debug("\tdelay: {0}",           main_config->rfid_conf.config->delay);
    spdlog::debug("database settings:");
    spdlog::debug("\thost: {0}",            main_config->db_conf.host);
    spdlog::debug("\tuser: {0}",            main_config->db_conf.user);
    spdlog::debug("\tpassword: {0}",        main_config->db_conf.passwd);
    spdlog::debug("\tdatabase: {0}",        main_config->db_conf.db);
    spdlog::debug("\tport: {0}",            main_config->db_conf.port);
    spdlog::debug("\tsocket: {0}",          main_config->db_conf.socket);
    spdlog::debug("\tclientflags: {0}",     main_config->db_conf.clientflags);
    spdlog::debug("hx711 settings:");
    spdlog::debug("\tdata pin: {0}",        main_config->hx_conf.data_pin);
    spdlog::debug("\tclock pin: {0}",       main_config->hx_conf.clock_pin);
    spdlog::debug("\trate: {0}",            main_config->hx_conf.rate);
    spdlog::debug("\tref unit: {0}",        main_config->hx_conf.ref_unit);
    spdlog::debug("\toffset: {0}",          main_config->hx_conf.offset);
    spdlog::debug("========== (debug) config file end   ==========");
}

int app_workspace::test_hx_samples() {
    if (this->hx711_controller.get() == nullptr) {
        set_err_screen_and_activate(get_localized_text("ERR_HX_ERROR"),
                                    get_localized_text("ERR_HX_NOT_INIT_DESC"));
        return 5;
    }

    if (!this->has_user()) {
        set_err_screen_and_activate(get_localized_text("ERR_NO_USER_LOGIN"),
                                    get_localized_text("ERR_NO_USER_LOGIN_DESC"));
        return 1; // can't measure if no user is logged in, this shouldn't be reachable
    }

    this->hx_measuring = true;
    this->test_hx_testing = true;
    this->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);

    this->measuring_thread.reset(new std::thread([]() {
        std::mutex mtx;
        mtx.lock();
        app_workspace *app_wrk = app_workspace::get_instance().get();

        // constant iterations
        for (int i = 0; i < HX_TEST_CONST_IT_COUNT; i++) {
            app_wrk->hx711_values_samples(HX_TEST_SAMP_COUNT);
            auto res = std::chrono::duration_cast<std::chrono::milliseconds>(app_wrk->m_end - app_wrk->m_start).count();
            app_wrk->test_hx_samples_times[i].first = res;
            app_wrk->test_hx_samples_times[i].second = 
                "Smps: " + std::to_string(HX_TEST_SAMP_COUNT) + " - " +  std::to_string(res) + " ms";
        }

        // increasing iterations
        for (int i = 0; i < HX_TEST_INC_IT_COUNT; i++) {
            app_wrk->hx711_values_samples(HX_TEST_SAMP_COUNT * (i + 1));
            auto res = std::chrono::duration_cast<std::chrono::milliseconds>(app_wrk->m_end - app_wrk->m_start).count();
            app_wrk->test_hx_samples_times[i + HX_TEST_CONST_IT_COUNT].first = res;
            app_wrk->test_hx_samples_times[i + HX_TEST_CONST_IT_COUNT].second = 
                "Smps: " + std::to_string(HX_TEST_SAMP_COUNT * (i + 1)) + " - " +  std::to_string(res) + " ms";
        }

        long worst_time = 0, best_time = 10000000;
        for (int i = 0; i < HX_TEST_CONST_IT_COUNT; i++) {
            spdlog::warn("app_config.cpp - test_hx_samples - samples: {0} - time {1} ms", 
                HX_TEST_SAMP_COUNT, app_wrk->test_hx_samples_times[i].first);
            if (app_wrk->test_hx_samples_times[i].first < best_time) best_time = app_wrk->test_hx_samples_times[i].first;
            if (app_wrk->test_hx_samples_times[i].first > worst_time) worst_time = app_wrk->test_hx_samples_times[i].first;
        }
        for (int i = 0; i < HX_TEST_INC_IT_COUNT; i++) {
            spdlog::warn("app_config.cpp - test_hx_samples - samples: {0} - time {1} ms", 
                HX_TEST_SAMP_COUNT * (i + 1), app_wrk->test_hx_samples_times[i + HX_TEST_CONST_IT_COUNT].first);
        }

        spdlog::warn("app_config.cpp - test_hx_samples - worst time - samples: {0} - time {1} ms",  
            HX_TEST_SAMP_COUNT, worst_time);
        spdlog::warn("app_config.cpp - test_hx_samples - best time - samples: {0} - time {1} ms",  
            HX_TEST_SAMP_COUNT, best_time);

        app_wrk->hx_measuring = false;
        app_wrk->test_hx_testing = false;
        app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
        mtx.unlock();
    }));

    this->measuring_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway

    return 0;
}

int app_workspace::test_hx_timeout() {
    if (this->hx711_controller.get() == nullptr) {
        set_err_screen_and_activate(get_localized_text("ERR_HX_ERROR"),
                                    get_localized_text("ERR_HX_NOT_INIT_DESC"));
        return 5;
    }

    if (!this->has_user()) {
        set_err_screen_and_activate(get_localized_text("ERR_NO_USER_LOGIN"),
                                    get_localized_text("ERR_NO_USER_LOGIN_DESC"));
        return 1; // can't measure if no user is logged in, this shouldn't be reachable
    }

    this->hx_measuring = true;
    this->test_hx_testing = true;
    this->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);

    this->measuring_thread.reset(new std::thread([]() {
        std::mutex mtx;
        mtx.lock();
        app_workspace *app_wrk = app_workspace::get_instance().get();

        // constant iterations
        for (int i = 0; i < HX_TEST_CONST_IT_COUNT; i++) {
            auto res = app_wrk->hx711_values_timeout(HX_TEST_TIME_COUNT);
            app_wrk->test_hx_timeout_collected[i].first = res.size();
            app_wrk->test_hx_timeout_collected[i].second = 
                "In: " + std::to_string(HX_TEST_TIME_COUNT) + "s - " +  std::to_string(res.size()) + " samples";
        }

        // increasing iterations
        for (int i = 0; i < HX_TEST_INC_IT_COUNT; i++) {
            auto res = app_wrk->hx711_values_timeout(HX_TEST_TIME_COUNT * (i + 1));
            app_wrk->test_hx_timeout_collected[i + HX_TEST_CONST_IT_COUNT].first = res.size();
            app_wrk->test_hx_timeout_collected[i + HX_TEST_CONST_IT_COUNT].second = 
                "In: " + std::to_string(HX_TEST_TIME_COUNT * (i + 1)) + "s - " +  std::to_string(res.size())+" samples";
        }

        long least_samples = 100000, most_samples = 0;
        for (int i = 0; i < HX_TEST_CONST_IT_COUNT; i++) {
            spdlog::warn("app_config.cpp - test_hx_timeout - timeout: {0} s - samples {1}", 
                HX_TEST_TIME_COUNT, app_wrk->test_hx_timeout_collected[i].first);
            if (app_wrk->test_hx_timeout_collected[i].first < least_samples) 
                least_samples = app_wrk->test_hx_timeout_collected[i].first;
            if (app_wrk->test_hx_timeout_collected[i].first > most_samples) 
                most_samples = app_wrk->test_hx_timeout_collected[i].first;
        }
        for (int i = 0; i < HX_TEST_INC_IT_COUNT; i++) {
            spdlog::warn("app_config.cpp - test_hx_timeout - timeout: {0} s - samples {1}", 
                HX_TEST_TIME_COUNT * (i + 1), app_wrk->test_hx_timeout_collected[i + HX_TEST_CONST_IT_COUNT].first);
        }

        spdlog::warn("app_config.cpp - test_hx_timeout - least samples - timeout: {0} s - samples {1}",  
            HX_TEST_TIME_COUNT, least_samples);
        spdlog::warn("app_config.cpp - test_hx_timeout - most samples - timeout: {0} s - samples {1}",  
            HX_TEST_TIME_COUNT, most_samples);

        app_wrk->hx_measuring = false;
        app_wrk->test_hx_testing = false;
        app_wrk->get_scr_mgr()->refresh_screen_elements(ADMIN_CTRL_SCREEN);
        mtx.unlock();
    }));

    this->measuring_thread->detach(); // not waiting for the thread to finish, but ignoring other events anyway

    return 0;
}

bool app_workspace::get_kb_testing() {
    return keyboard::get_kb_testing_inner();
}

void app_workspace::set_kb_testing(bool val) {
    keyboard::set_kb_testing_inner(val);
}


/*********************************************************************************************************************/
/*       Functions used by app_wrokspace for hx711 thread measuring                                                  */
/*********************************************************************************************************************/

void threaded_values_read() {
    std::mutex mtx;
    std::vector<HX711::Value> res;
    
    mtx.lock();
    app_workspace *app_wrk = app_workspace::get_instance().get();

    if (app_wrk->selected_hx_measure_opt == 0) {
        res = app_wrk->hx711_values_samples(app_wrk->samples_in);
    } else if (app_wrk->selected_hx_measure_opt == 1) {
        res = app_wrk->hx711_values_timeout(app_wrk->timeout_in);
    }

    // thanks to this measuring isn't save when cancelled mid-way
    /*
    if (app_wrk->hx_not_finished_yet) {
        app_wrk->hx_measuring = false;
        app_wrk->hx_continuous = false;
        app_wrk->hx_not_finished_yet = false;
        app_wrk->refresh_current_screen = true;
        mtx.unlock();
        return; // could use goto end of this function
    }
    */

    if (res.size() <= 0) {
        app_wrk->set_err_screen_and_activate(get_localized_text("ERR_S2_MEASURING_FAILURE"),
                                    get_localized_text("ERR_S2_MEASURING_NO_VALS"));
        return;
    }
    
    unsigned long m_len = 
        std::chrono::duration_cast<std::chrono::microseconds>(app_wrk->m_end - app_wrk->m_start).count();
    std::time_t start_timet = std::chrono::system_clock::to_time_t(app_wrk->m_start);
    std::time_t end_timet = std::chrono::system_clock::to_time_t(app_wrk->m_end);
    std::tm start_tm, end_tm;
    localtime_r(&start_timet, &start_tm);
    localtime_r(&end_timet, &end_tm);

    std::vector<double> values = app_wrk->normalize_raw_values(res);
    app_wrk->userspace->init_measured(values, &start_tm, &end_tm, m_len); // inits userspace->measured
    app_wrk->userspace->get_last_hx_measuring()->log_measurement_to_debug();

    // insert to db
    int rv = app_wrk->db_conn->insert_measurement(app_wrk->userspace->measured.get());
    if (rv) {
        spdlog::error("app_workspace.cpp - hx711_measurement: failet to insert measurement to DB.");
    } else {
        rv = app_wrk->db_conn->increment_user_measurement_count(app_wrk->userspace->get_measured_id());
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Successfully inserted measurement but"
                          " failed to increment user measurement_count.");
        } else {
            // to save reloading from db
            app_wrk->userspace->get_measured_user()->measure_count++;
        }

        rv = app_wrk->db_conn->query_measurement_headers(app_wrk->userspace->get_measured_user());
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve (update)"
                          " user measurement headers.");
        } else {
            app_wrk->get_scr_mgr()->refresh_screen_elements(3); // refresh measurement list
        }

        // fetch username of measuree and measurer (last measurement preview on screen 2)
        std::string uname;
        rv = app_wrk->db_conn->query_username(uname, app_wrk->userspace->measured->measurer_id);
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measurer username.");
        } else {
            if (app_wrk->userspace->measured->measurer_id != app_wrk->userspace->measured->measuree_id) {
                std::string sub_uname;
                rv = app_wrk->db_conn->query_username(sub_uname, app_wrk->userspace->measured->measuree_id);

                if (rv) {
                    spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measuree username.");
                } else {
                    app_wrk->userspace->measured->init_convinience_vars(uname.c_str(), sub_uname.c_str());
                }
            } else {
                app_wrk->userspace->measured->init_convinience_vars(uname.c_str(), uname.c_str());
            }
        }
        // else { // this branch is for debug
        //     measurement_header *h = app_wrk->userspace->get_user()->measur_headers.at(0).get();
        //     measurement *m = new measurement();
        //     rv = app_wrk->db_conn->query_measurement(m, h->id);
        //     if (rv) {
        //         spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measurement that was just inserted");
        //     } else {
        //         app_wrk->userspace->measured->compare(m);
        //     }
        // }
    }

    // following 2 flags are reset just to be sure
    app_wrk->hx_continuous = false;
    app_wrk->hx_not_finished_yet = false;

    app_wrk->hx_measuring = false;
    app_wrk->refresh_current_screen = true;

    mtx.unlock();
}

void threaded_tare() {
    std::mutex mtx;
    std::vector<HX711::Value> res;
    
    mtx.lock();
    app_workspace *app_wrk = app_workspace::get_instance().get();

    spdlog::info("app_workspace.cpp - Starting taring");
    app_wrk->s6_tare_complete = false; // making sure tare is false

    // tare start
    HX711::Value orig_offset = app_wrk->hx711_controller->getOffset();
    HX711::Value orig_ref_unit = app_wrk->hx711_controller->getReferenceUnit();

    try {
        app_wrk->hx711_controller->setReferenceUnit(1);
        app_wrk->hx711_controller->setOffset(0);
    } catch (std::exception &ex) {
        spdlog::error("app_workspace.cpp - Caught exception while taring HX711");
        app_wrk->hx_measuring = false;
        app_wrk->refresh_current_screen = true;

        mtx.unlock();
        return;
    }

    size_t samples = app_wrk->s6_samples_in;
    app_wrk->zero_value = app_wrk->hx711_controller->read(HX711::Options(samples));

    try {
        app_wrk->hx711_controller->setReferenceUnit(orig_ref_unit);
        app_wrk->hx711_controller->setOffset(orig_offset);
    } catch (std::exception &ex) {
        spdlog::error("app_workspace.cpp - Caught exception while cleaning after HX711 calibration");
        app_wrk->hx_measuring = false;
        app_wrk->refresh_current_screen = true;

        mtx.unlock();
        return;
    }
    // tare end

    spdlog::info("app_workspace.cpp - Successfully finished calibrating");
    app_wrk->hx_measuring = false;
    app_wrk->s6_tare_complete = true;
    app_wrk->refresh_current_screen = true;

    mtx.unlock();
}

void threaded_calibration() {
    std::mutex mtx;
    std::vector<HX711::Value> res;
    
    mtx.lock();
    app_workspace *app_wrk = app_workspace::get_instance().get();

    spdlog::info("app_workspace.cpp - Starting calibration");

    // calibration start
    HX711::Value orig_offset = app_wrk->hx711_controller->getOffset();
    HX711::Value orig_ref_unit = app_wrk->hx711_controller->getReferenceUnit();

    try {
        app_wrk->hx711_controller->setReferenceUnit(1);
        app_wrk->hx711_controller->setOffset(0);
    } catch (std::exception &ex) {
        spdlog::error("app_workspace.cpp - Caught exception while preparing for HX711 calibration");
        app_wrk->hx_measuring = false;
        app_wrk->refresh_current_screen = true;

        mtx.unlock();
        return;
    }


    double known_weight = app_wrk->s6_ref_mass_in;
    size_t samples = app_wrk->s6_samples_in;

    const double raw = app_wrk->hx711_controller->read(HX711::Options(samples));
    const double ref_unit_float = (raw - app_wrk->zero_value) / known_weight;
    app_wrk->ref_unit = static_cast<HX711::Value>(round(ref_unit_float));

    if (app_wrk->ref_unit == 0)
        app_wrk->ref_unit = 1;

    // FIXME remove s6_calib...  those are useless, same data stored twice (but s6_calib.. is used in gui)
    // offset
    app_wrk->s6_calib_offset = app_wrk->zero_value;

    // ref_unit
    app_wrk->s6_calib_ref_unit = app_wrk->ref_unit;

    try {
        app_wrk->hx711_controller->setReferenceUnit(orig_ref_unit);
        app_wrk->hx711_controller->setOffset(orig_offset);
    } catch (std::exception &ex) {
        spdlog::error("app_workspace.cpp - Caught exception while cleaning after HX711 calibration");
        app_wrk->hx_measuring = false;
        app_wrk->refresh_current_screen = true;

        mtx.unlock();
        return;
    }

    // calibration end

    spdlog::info("app_workspace.cpp - Successfully finished calibrating");
    app_wrk->s6_calibration_finished = true;
    app_wrk->s6_tare_complete = false;
    app_wrk->hx_measuring = false;
    app_wrk->refresh_current_screen = true;

    mtx.unlock();
}

void threaded_values_read_looped() {
    std::mutex mtx;
    long failed_measures = 0, failed_inserts = 0; // in case measuring or insert fails, the number of fails will be logged
    unsigned long measure_count = 1;
    int consec_fm = 0, consec_fi = 0; // consecutive fail counter, after reaching several consecutive fails, end measuring
    int rv;
    std::vector<HX711::Value> res;
    std::unique_ptr<measurement> measr; // using own measurement instead of userspace->measured

    mtx.lock();
    app_workspace *app_wrk = app_workspace::get_instance().get();
    mtx.unlock();

    while (1) {
        mtx.lock();
        if (!app_wrk->hx_continuous || app_wrk->hx_not_finished_yet) break;
        mtx.unlock();

        // prepare next measuring iteration
        if (consec_fi > HX_CONT_MAX_CONSEC_FAIL_FI || consec_fm > HX_CONT_MAX_CONSEC_FAIL_FM) {
            spdlog::error("app_workspace.cpp - countinous measuring failed. Reason is because of consecutive"
                "measuring or insert failures. Failed measures: {0} (limit: {1}), failed inserts: {2} (limit: {3})",
                consec_fm, HX_CONT_MAX_CONSEC_FAIL_FI, consec_fi, HX_CONT_MAX_CONSEC_FAIL_FM);
            break;
        }
        measr.reset(new measurement());

        // read hx711 and get results
        std::vector<HX711::Value> res = app_wrk->hx711_values_samples(HX_CONT_SAMPLES);

        // was using this to avoid locking the whole thread while measuring, but moved this to hx711_values_samples()
        // mtx.lock();
        // app_wrk->m_start = std::chrono::high_resolution_clock::now();
        // mtx.unlock();
        // std::vector<HX711::Value> res = app_wrk->hx711_controller->getValues(HX_CONT_SAMPLES);
        // mtx.lock();
        // app_wrk->m_end = std::chrono::high_resolution_clock::now();
        // mtx.unlock();

        // stop measuring on request (hx_not_finished_yet), but don't break if there is data to write
        if (app_wrk->hx_not_finished_yet && res.size() <= 0) break;

        // check if results were measured (maybe can do size() != CONTINOUS_M_SAMPLES)
        if (res.size() <= 0) {
            failed_measures++;
            consec_fm++;
            continue; // failed measuring, try again
        }
        consec_fm = 0; // probably more effective than else consec_fm = 0; ?

        mtx.lock();
        unsigned long m_len = 
            std::chrono::duration_cast<std::chrono::microseconds>(app_wrk->m_end - app_wrk->m_start).count();
        std::time_t start_timet = std::chrono::system_clock::to_time_t(app_wrk->m_start);
        std::time_t end_timet = std::chrono::system_clock::to_time_t(app_wrk->m_end);
        std::tm start_tm, end_tm;
        localtime_r(&start_timet, &start_tm);
        localtime_r(&end_timet, &end_tm);

        std::vector<double> values = app_wrk->normalize_raw_values(res);
        measurement::init_measuring(measr.get(), app_wrk->get_userspace(), values, &start_tm, &end_tm, m_len);

        measr->measurement_part = measure_count;
        measure_count++;

        // db insert
        auto i_start = std::chrono::steady_clock::now();
        rv = app_wrk->db_conn->insert_measurement(measr.get());
        auto i_end = std::chrono::steady_clock::now();
        spdlog::debug("app_workspace.cpp - hx711_continuous_measurement: inserting measurement to DB"
            " took {0}ms ({1} us)", std::chrono::duration_cast<std::chrono::milliseconds>(i_end - i_start).count(),
                                    std::chrono::duration_cast<std::chrono::microseconds>(i_end - i_start).count());
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_continuous_measurement: failed to insert measurement to DB.");
            failed_inserts++;
            consec_fi++;
        }
        else consec_fi = 0; // not using continue, need else

        mtx.unlock();
    }

    // update measurement_number and screen 3 header list
    if ((measure_count - failed_inserts) > 1) { // atleast one result was measured and inserted
        rv = app_wrk->db_conn->increment_user_measurement_count(app_wrk->userspace->get_measured_id());
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Successfully inserted measurement but"
                          " FAILED to increment user measurement_count.");
        } else {
            // to save reloading from db
            app_wrk->userspace->get_measured_user()->measure_count++;
        }

        rv = app_wrk->db_conn->query_measurement_headers(app_wrk->userspace->get_measured_user());
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve (update)"
                          " user measurement headers.");
        } else {
            app_wrk->get_scr_mgr()->refresh_screen_elements(3); // refresh measurement list
        }

        // fetch username of measuree and measurer (convenience vars)
        // last measuring result (last part of continuous measuring)
        app_wrk->userspace->measured = std::unique_ptr(std::move(measr));

        std::string uname;
        rv = app_wrk->db_conn->query_username(uname, app_wrk->userspace->measured->measurer_id);
        if (rv) {
            spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measurer username.");
        } else {
            if (app_wrk->userspace->measured->measurer_id != app_wrk->userspace->measured->measuree_id) {
                std::string sub_uname;
                rv = app_wrk->db_conn->query_username(sub_uname, app_wrk->userspace->measured->measuree_id);

                if (rv) {
                    spdlog::error("app_workspace.cpp - hx711_measurement: Failed to retrieve measuree username.");
                } else {
                    app_wrk->userspace->measured->init_convinience_vars(uname.c_str(), sub_uname.c_str());
                }
            } else {
                app_wrk->userspace->measured->init_convinience_vars(uname.c_str(), uname.c_str());
            }
        }
        
    }

    // mtx.lock();
    app_wrk->hx_measuring = false;
    app_wrk->hx_continuous = false;
    app_wrk->hx_not_finished_yet = false;
    app_wrk->refresh_current_screen = true;
    mtx.unlock();
}

void threaded_measuring_watchdog(bool *msring, std::vector<HX711::Value> **watcher) {
    std::mutex mtx;
    int fail_cnt = 0;

    mtx.lock();
    app_workspace *app_wrk = app_workspace::get_instance().get();
    mtx.unlock();
        
    size_t val_len = 0;
    spdlog::debug("STARTING measuring watchdog");

    while (1) {
        mtx.lock();
        if (!(*msring)) break;

        if (*watcher != nullptr && ((*watcher)->size() - val_len) > 50) { // every X samples do something
            /*
            auto tmp_start = (*watcher)->begin() + val_len; // iterator start (not doing -1, because not including last)
            auto tmp_end = (*watcher)->end(); // iterator end
            std::vector<HX711::Value> tmp(tmp_start, tmp_end);
            */

            if (val_len < (*watcher)->size()) {
                std::vector<HX711::Value> tmp;
                for (size_t i = val_len; i < (*watcher)->size(); i++) {
                    tmp.push_back((*watcher)->at(i));
                }
                val_len = (*watcher)->size();

                auto norm_values = app_wrk->normalize_raw_values(tmp);
                app_wrk->observed_measurement.reset(new measurement);
                measurement::init_simple_measuring(app_wrk->observed_measurement.get(), norm_values);
                app_wrk->refresh_current_screen = true;
            } else
                fail_cnt++;

            if (fail_cnt == 20) { // 20 * 50 takes 1 second to finish - maybe make it quicker (< 20)
                printf("breaking on fail_cnt\n");
                *msring = false;
                break; // after 20 "not update" break
            }
        }
        mtx.unlock();

        std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }

    spdlog::debug("ENDING measuring watchdog");
}